/*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */

package org.wso2.esb;

import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.deployment.Deployer;
import org.apache.axis2.deployment.DeploymentClassLoader;
import org.apache.axis2.deployment.DeploymentException;
import org.apache.axis2.deployment.repository.util.DeploymentFileData;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.synapse.config.xml.MediatorFactory;
import org.apache.synapse.config.xml.MediatorFactoryFinder;
import org.apache.synapse.config.xml.MediatorSerializer;
import org.apache.synapse.config.xml.MediatorSerializerFinder;

import java.io.*;
import java.net.URL;

/**
 * This will support the hot deployment and hot update of the mediators at runtime using the
 * Axis2 concepts of deployers
 */
public class MediatorDeployer implements Deployer {

    /**
     * Holds the log variable for logging purposes
     */
    private static final Log log = LogFactory.getLog(MediatorDeployer.class);

    /**
     * ConfigurationContext of Axis2
     */
    private ConfigurationContext cfgCtx = null;

    /**
     * Initializes the Deployer
     *
     * @param configurationContext - ConfigurationContext of Axis2 from which
     *                             the deployer is initialized
     */
    public void init(ConfigurationContext configurationContext) {
        this.cfgCtx = configurationContext;
    }

    /**
     * This will be called when there is a change in the specified deployement
     * folder (in the axis2.xml) and this will load the relevant classe to the system and
     * registeres them with the MediatorFactoryFinder
     *
     * @param deploymentFileData - describes the updated file
     * @throws DeploymentException - in case an error on the deployment
     */
    public void deploy(DeploymentFileData deploymentFileData) throws DeploymentException {

        log.info("Loading mediator from: " + deploymentFileData.getAbsolutePath());

        // get the context class loader for the later restore of the context class loader
        ClassLoader prevCl = Thread.currentThread().getContextClassLoader();

        try {

            DeploymentClassLoader urlCl = new DeploymentClassLoader(
                    new URL[]{deploymentFileData.getFile().toURL()}, null, prevCl);
            Thread.currentThread().setContextClassLoader(urlCl);

            // MediatorFactory registration
            URL facURL = urlCl.findResource(
                    "META-INF/services/org.apache.synapse.config.xml.MediatorFactory");
            if (facURL != null) {
                InputStream facStream = facURL.openStream();
                InputStreamReader facreader = new InputStreamReader(facStream);

                StringBuffer facSB = new StringBuffer();
                int c;
                while ((c = facreader.read()) != -1) {
                    facSB.append((char) c);
                }

                String[] facClassName = facSB.toString().split("\n");
                for (int i = 0; i < facClassName.length; i++) {
                    log.info("Registering the Mediator factory: " + facClassName[i]);
                    Class facClass = urlCl.loadClass(facClassName[i]);
                    MediatorFactory facInst = (MediatorFactory) facClass.newInstance();
                    MediatorFactoryFinder.getInstance()
                            .getFactoryMap().put(facInst.getTagQName(), facClass);
                    log.info("Mediator loaded and registered for " +
                            "the tag name: " + facInst.getTagQName());
                    loadUIComponents(facInst.getTagQName().getLocalPart(), urlCl);
                }
            } else {
                handleException("Unable to find the MediatorFactory implementation. " +
                        "Unable to register the MediatorFactory with the FactoryFinder");
            }

            // MediatorSerializer registration
            URL serURL = urlCl.findResource(
                    "META-INF/services/org.apache.synapse.config.xml.MediatorSerializer");
            if (serURL != null) {
                InputStream serStream = serURL.openStream();
                InputStreamReader serReader = new InputStreamReader(serStream);

                StringBuffer serSB = new StringBuffer();
                int c;
                while ((c = serReader.read()) != -1) {
                    serSB.append((char) c);
                }

                String[] serClassName = serSB.toString().split("\n");
                for (int i = 0; i < serClassName.length; i++) {
                    log.info("Registering the Mediator serializer: " + serClassName[i]);
                    Class serClass = urlCl.loadClass(serClassName[i]);
                    MediatorSerializer serInst = (MediatorSerializer) serClass.newInstance();
                    MediatorSerializerFinder.getInstance()
                            .getSerializerMap().put(serInst.getMediatorClassName(), serInst);
                    log.info("Mediator loaded and registered for " +
                            "the serialization as: " + serInst.getMediatorClassName());
                }
            } else {
                handleException("Unable to find the MediatorSerializer implementation. " +
                        "Unable to register the MediatorSerializer with the SerializerFinder");

            }

        } catch (IOException e) {
            handleException("I/O error in reading the mediator jar file", e);
        } catch (ClassNotFoundException e) {
            handleException("Unable to find the specified class on the path or in the jar file", e);
        } catch (IllegalAccessException e) {
            handleException("Unable to load the class from the jar", e);
        } catch (InstantiationException e) {
            handleException("Unable to instantiate the class specified", e);
        } finally {
            // restore the class loader back
            if (log.isDebugEnabled()) {
                log.debug("Restoring the context class loader to the original");
            }
            Thread.currentThread().setContextClassLoader(prevCl);
        }
    }

    private void loadUIComponents(String tagName, DeploymentClassLoader clsL) {

        // finding the root directory for the UI stuff todo
        String rootUIDir = "/home/ruwan/dev/esb/modules/distribution/target/wso2esb-SNAPSHOT/webapp/extensions/core/";

        // loading the view XSLT
        loadFileFromJar("www/xslt/" + tagName + "_view.xsl",
                rootUIDir + "xslt/mediators/" + tagName + "_view.xsl", clsL);

        // loading the edit XSLT
        loadFileFromJar("www/xslt/" + tagName + "_edit.xsl",
                rootUIDir + "xslt/mediators/" + tagName + "_edit.xsl", clsL);

        // loading the data collecting js
        loadFileFromJar("www/js/" + tagName + "_mediator.js",
                rootUIDir + "js/mediators/" + tagName + "_mediator.js", clsL);

        // appending the view XSL to the includes file
        File includesXSL = new File(rootUIDir + "xslt/mediators/custom_mediator_includes.xsl");
        includesXSL.deleteOnExit();
        try {
            StringBuffer buff = new StringBuffer();
            InputStreamReader srcReader = new InputStreamReader(includesXSL.toURL().openStream());

            int c;
            while ((c = srcReader.read()) != -1) {
                buff.append((char) c);
            }
            String initInc = buff.toString();
            if (!initInc.contains("<xsl:include href=" + tagName + "_view.xsl />")) {
                int breakPoint = initInc.indexOf("/>", initInc.lastIndexOf("<xsl:include ")) + 2;
                String modifiedInc = initInc.substring(0, breakPoint) + "\n\t<xsl:include href=\"" + tagName + "_view.xsl\" />" + initInc.substring(breakPoint);
                BufferedWriter out = new BufferedWriter(new FileWriter(includesXSL));
                out.write(modifiedInc);
                out.flush();
                out.close();
            }
        } catch (IOException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }

        // appending the data collecting js script file to the js.html
        File jsHtml = new File(rootUIDir + "js.html");
        try {
            StringBuffer buff = new StringBuffer();
            InputStreamReader srcReader = new InputStreamReader(jsHtml.toURL().openStream());

            int c;
            while ((c = srcReader.read()) != -1) {
                buff.append((char) c);
            }
            if (!buff.toString().contains("<script language=\"javascript\" src=\"extensions/core/js/mediators/" + tagName + "_mediator.js\"></script>")) {
                BufferedWriter out = new BufferedWriter(new FileWriter(jsHtml));
                out.write(buff.toString() + "\n<script language=\"javascript\" src=\"extensions/core/js/mediators/" + tagName + "_mediator.js\"></script>");
                out.flush();
                out.close();
            }
        } catch (IOException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
    }

    private void loadFileFromJar(String srcName, String dest, DeploymentClassLoader clsL) {

        try {
            URL srcURL = clsL.findResource(srcName);
            File destFile = new File(dest);
            if (!destFile.exists()) {
                destFile.createNewFile();
            }
            StringBuffer buff = new StringBuffer();
            InputStreamReader srcReader = new InputStreamReader(srcURL.openStream());
            int c;
            while ((c = srcReader.read()) != -1) {
                buff.append((char) c);
            }
            BufferedWriter out = new BufferedWriter(new FileWriter(destFile));
            out.write(buff.toString());
            out.flush();
            out.close();
        } catch (IOException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
    }

    /**
     * This will not be implemented because we do not support changing the directory at runtime
     *
     * @param string -
     */
    public void setDirectory(String string) {
        // we do not support changing the directory
    }

    /**
     * This will not be implemented because we do not support changing the extension at runtime
     *
     * @param string -
     */
    public void setExtension(String string) {
        // we do not support changing the extension
    }

    /**
     * This will be called when a particulr jar file is deleted from the specified folder
     *
     * @param string - filename of the deleted file
     * @throws DeploymentException - incase of an error in undeployment
     */
    public void unDeploy(String string) throws DeploymentException {
        // todo: implement the undeployement
    }

    private void handleException(String message, Exception e) throws DeploymentException {
        if (log.isDebugEnabled()) {
            log.debug(message, e);
        }
        throw new DeploymentException(message, e);
    }

    private void handleException(String message) throws DeploymentException {
        if (log.isDebugEnabled()) {
            log.debug(message);
        }
        throw new DeploymentException(message);
    }
}
