/*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */

package org.wso2.esb;

import org.apache.axiom.om.OMElement;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.utils.ServerConfiguration;
import org.wso2.utils.ServerConfigurationException;

import java.io.File;

/**
 * Loads the service bus configuration from the config file server.xml and manages it
 */

public class ServiceBusConfiguration {

    private static Log log = LogFactory.getLog(ServiceBusConfiguration.class);
    private final static String DEF_LOCATION = ServiceBusConstants.ESB_CONF_DIRECTORY  + "server.xml";

    private static ServiceBusConfiguration configuration = null;
    private ServerConfiguration serverConfiguration = null;

    private String esbHome = null;
    private String serverXml = null;
    private String synapseXml = null;
    private String axis2Xml = null;
    private String axis2Repo = null;
    private String resolveRoot = null;
    private String serverName = null;
    private String jmxAgentName = ServiceBusConstants.DEFAULT_JMX_AGENT_NAME;
    private int rmiPort = 1099;

    private ServiceBusConfiguration() {}

    public static ServiceBusConfiguration getInstance() {
        if (configuration == null) {
            configuration = new ServiceBusConfiguration();
        }
        return configuration;
    }

    /**
     * Load the server XML, requires the esbHome to be valid
     */
    private void loadServerXML() {
        if (esbHome == null) {
            if (System.getProperty(ServiceBusConstants.ESB_HOME) == null) {
                esbHome = ".";
            } else {
                esbHome = System.getProperty(ServiceBusConstants.ESB_HOME);
            }
        }

        try {
            serverConfiguration = ServerConfiguration.getInstance();
            serverConfiguration.init(DEF_LOCATION);
            log.info("Configuration loaded from : " + DEF_LOCATION);

        } catch (ServerConfigurationException e) {
            log.error("Error initalizing ServiceBusConfiguration using : " + DEF_LOCATION);
        }
    }

    public String getFirstProperty(String key) {
        if (serverConfiguration == null) {
            loadServerXML();
        }
        return serverConfiguration.getFirstProperty(key);
    }

    public int getFirstPropertyAsInt(String key, int def) {
        if (serverConfiguration == null) {
            loadServerXML();
        }

        int ret = def;
        String prop = serverConfiguration.getFirstProperty(key);
        if (prop != null) {
            try {
                ret = Integer.parseInt(prop);
            } catch (NumberFormatException ignore) {
                log.warn("Invalid property '" + key + "' in server.xml Using default : " + def);
            }
        }
        return ret;
    }

    public OMElement getDocumentElement() {
        if (serverConfiguration == null) {
            loadServerXML();
        }
        return serverConfiguration.getDocumentElement();
    }

    private void handleException(String msg) {
        log.error(msg);
        throw new ServiceBusException(msg);
    }

    //-- getters and setters --
    public String getEsbHome() {
        return esbHome;
    }

    public void setEsbHome(String esbHome) {
        if (!new File(esbHome).isAbsolute()) {
            this.esbHome = new File(esbHome).getAbsolutePath();
        } else {
            this.esbHome = esbHome;
        }
    }

    public String getServerXml() {
        return serverXml;
    }

    public void setServerXml(String serverXml) {
        this.serverXml = serverXml;
    }

    public String getSynapseXml() {
        return synapseXml;
    }

    public void setSynapseXml(String synapseXml) {
        this.synapseXml = synapseXml;
    }

    public String getAxis2Xml() {
        return axis2Xml;
    }

    public void setAxis2Xml(String axis2Xml) {
        this.axis2Xml = axis2Xml;
    }

    public String getAxis2Repo() {
        return axis2Repo;
    }

    public void setAxis2Repo(String axis2Repo) {
        this.axis2Repo = axis2Repo;
    }

    public String getResolveRoot() {
        return resolveRoot;
    }

    public void setResolveRoot(String resolveRoot) {
        this.resolveRoot = resolveRoot;
    }

    public String getJmxAgentName() {
        return jmxAgentName;
    }

    public int getConsolePort() {
        return getFirstPropertyAsInt("AdminConsole.Port", 9444);
    }

    public int getHttpPort() {
        return getFirstPropertyAsInt("HTTPPort", 8280);
    }

    public int getHttpsPort() {
        return getFirstPropertyAsInt("HTTPSPort", 4433);
    }

    public int getDerbyPort() {
        return getFirstPropertyAsInt("DerbyPort", 1528);
    }

    public boolean startEmbeddedDerby() {
        return Boolean.valueOf(getFirstProperty("StartEmbeddedDerby"));
    }

    public int getRmiPort() {
        return rmiPort;
    }

    public void setRmiPort(int rmiPort) {
        this.rmiPort = rmiPort;
    }

    public String getContextPath() {
        return getFirstProperty("ContextPath");
    }

    public String getServerName() {
        return serverName;
    }

    public void setServerName(String serverName) {
        this.serverName = serverName;
    }
}
