/*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */

package org.wso2.esb;


import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.xpath.AXIOMXPath;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.description.TransportInDescription;
import org.apache.axis2.description.TransportOutDescription;
import org.apache.axis2.transport.TransportListener;
import org.apache.axis2.transport.TransportSender;
import org.apache.axis2.AxisFault;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.derby.drda.NetworkServerControl;
import org.apache.synapse.ServerManager;
import org.apache.synapse.SynapseConstants;
import org.apache.synapse.transport.base.ManagementSupport;
import org.apache.synapse.config.SynapseConfiguration;
import org.apache.synapse.core.SynapseEnvironment;
import org.apache.synapse.registry.Registry;
import org.apache.log4j.PropertyConfigurator;
import org.jaxen.JaxenException;
import org.jaxen.SimpleNamespaceContext;
import org.jaxen.XPath;
import org.wso2.adminui.UIProcessingException;
import org.wso2.adminui.UIProcessor;
import org.wso2.esb.jmx.ESBJMXAuthenticator;
import org.wso2.esb.jmx.mbean.*;
import org.wso2.esb.registry.ESBRegistry;
import org.wso2.esb.statistics.StatisticsReporterThread;
import org.wso2.esb.util.HibernateConfigCache;
import org.quartz.impl.DirectSchedulerFactory;
import org.quartz.SchedulerException;
import org.quartz.Scheduler;

import javax.management.*;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;
import javax.management.remote.JMXAuthenticator;
import javax.xml.namespace.QName;
import java.io.File;
import java.io.IOException;
import java.lang.management.ManagementFactory;
import java.rmi.NoSuchObjectException;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.UnicastRemoteObject;
import java.util.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.net.InetAddress;

/*
 * The Service bus manager
 */
public class ServiceBusManager {

    private static final Log log = LogFactory.getLog(ServiceBusManager.class);

    private static final String JMX_PREFIX = "service:jmx:rmi:///jndi/rmi://";
    private static final QName USERNAME_QN = new QName("http://www.wso2.org/projects/esb", "Username");
    private static final QName PASSWORD_QN = new QName("http://www.wso2.org/projects/esb", "Password");

    /** Singleton reference to itself */
    private static ServiceBusManager instance;
    /* The thread which reads statistics from memory periodically and write them to the DB */
    private StatisticsReporterThread statisticsReporterThread;
    /** JMX server*/
    private JMXConnectorServer jmxConnectorServer;
    /** Embedded Derby server */
    private NetworkServerControl networkServerControl;
    /** RMI Registry for JMX */
    private java.rmi.registry.Registry rmiRegistry;

    /** Underlying Axis2 configuration context */
    private ConfigurationContext configurationContext;
    /** Has the ESB been already initialized? */
    private boolean alreadyInitialized = false;
    /** The ServiceBusConfiguration driving the instance */
    private ServiceBusConfiguration sbc = ServiceBusConfiguration.getInstance();
    /** Save the TCCL of the initial thread that starts the ESB for future use. When JMX calls are
     * received via RMI connections, re-start etc may otherwise fail due to classloading issues */
    private ClassLoader classLoader = null;
    /** Is the service bus manager (i.e. this) started? */
    private boolean serviceBusManagerStarted = false;

    private ServiceBusManager() {}

    public static ServiceBusManager getInstance() {
        if (instance == null) {
            instance = new ServiceBusManager();
        }
        return instance;
    }

    /**
     * Perform one time configuration for the JVM such as Log4J initialization, and core
     * JMX Setup of the ServerView MBean
     */
    public void init() {
        if (alreadyInitialized) {
            return;
        }

        // save web app classloader
        classLoader = Thread.currentThread().getContextClassLoader();

        // initialize log4j
        URL log4jURL = null;
        try {
            log4jURL = new URL("file:" + sbc.getEsbHome() + File.separator
                + ServiceBusConstants.ESB_CONF_DIRECTORY + ServiceBusConstants.LOG4J_PROPS);
            PropertyConfigurator.configure(log4jURL);
        } catch (MalformedURLException ignore) {
            System.out.println("Log4J configuration initialization failure. Using : " + log4jURL);
        }

        // register a JVM shutdown hook
        Thread shutdownHook = new Thread() {
            public void run() {
                if (serviceBusManagerStarted) {
                    shutdown();
                }
            }
        };
        shutdownHook.setName("WSO2 ESB Shutdown Hook");
        Runtime.getRuntime().addShutdownHook(shutdownHook);

        HibernateConfigCache.clearCache();

        // start JMX
        startJMXService();
        registerMBean(
            ManagementFactory.getPlatformMBeanServer(), new ServiceBusControl(),
            sbc.getJmxAgentName() + ":Type=Admin,ConnectorName=ServerAdmin");
        // ask Synapse transports to use the ESB agent name
        System.setProperty(ServiceBusConstants.JMX_AGENT_NAME, sbc.getJmxAgentName());

        alreadyInitialized = true;
        serviceBusManagerStarted = true;
    }

    /**
     * Start the WSO2 ESB. This method can be invoked multiple times within a JVM to re-start
     * the ESB after an invocation of the stop() method. However, a shutdown() would terminate
     * the core ServerViewMBean and will prevent any further start() invocations over JMX
     */
    public void start() {
        if (!alreadyInitialized) {
            // if one time (per JVM) initialization has not yet taken place, do it..
            init();
        }

        log.info("Starting the WSO2 Enterprise Service Bus..");

        // start embedded Derby database server
        if (sbc.startEmbeddedDerby()) {
            startDataBaseServer();
        }

        // start Synapse
        ServerManager.getInstance().setSynapseHome(sbc.getEsbHome());
        ServerManager.getInstance().setAxis2Repolocation(sbc.getAxis2Repo());
        ServerManager.getInstance().setAxis2Xml(sbc.getAxis2Xml());        
        ServerManager.getInstance().setSynapseXMLPath(sbc.getSynapseXml());
        ServerManager.getInstance().setResolveRoot(sbc.getResolveRoot());
        ServerManager.getInstance().setServerName(sbc.getServerName());
        ServerManager.getInstance().start();

        configurationContext = ServerManager.getInstance().getConfigurationContext();
        // set ESB defaults to the Synapse config (e.g. Registry)
        setESBDefaults(configurationContext);

        // start the statistics reporter
        startStatisticsReporter();

        // generate UI pages and start JMX
        generateUIPages();
        registerMBeans();
        log.info("[ESB] Start request completed");
    }

    /**
     * Stops the ESB, and allow it to be re-started via the start() method subsequently. This
     * will stop the Synapse engine (and transports) and also unregister all JMX MBeans, leaving
     * only the core ServerView MBean that could be used to restart the ESB
     * @throws ServiceBusException
     */
    public void stop() throws ServiceBusException {
        log.info("Stopping the WSO2 Enterprise Service Bus..");
        stopStatisticsReporter();
        stopDataBaseServer();
        ServerManager.getInstance().stop();
        unregisterMBeans();
        log.info("[ESB] Stop request completed");
    }

    /**
     * Stop the ESB and the ServerView MBean and JMX control, and shutdown the JVM
     */
    public void shutdown() {
        stop();
        stopJMXService();
        log.info("[ESB] Shutdown completed");
        serviceBusManagerStarted = false;

        // wait a few seconds more just to be really sure..
        try {
            log.info("[ESB] System will exit in approximately 3 seconds..");
            Thread.sleep(3000);
        } catch (InterruptedException ignore) {}

        // if this was invoked by the shutdown hook, we should not call System.exit
        if (!Thread.currentThread().getName().endsWith("Hook")) {
            System.exit(0);
        }
    }

    /**
     * Graceful shutdown of the ESB
     *
     * Put all transports that can go into maintenence mode into maintenence
     * wait for all transport listener thread pools to reach zero
     *
     * @param waitMillis the maximum number of ms to wait until a graceful shutdown is achieved,
     * before forcing a shutdown
     */
    public void shutdownGracefully(long waitMillis) {

        long endTime = System.currentTimeMillis() + waitMillis;
        log.info("Requesting a graceful shutdown at : " + new Date() + " in a maximum of : " +
            (waitMillis/1000) + " seconds");

        // put transports into maintenence
        startMaintenance();

        // wait till listeners are idle
        boolean safeToShutdown = false;

        while (!safeToShutdown) {

            int pendingThreads = 0;
            Map<String, TransportInDescription> trpIns =
                configurationContext.getAxisConfiguration().getTransportsIn();
            Map<String, TransportOutDescription> trpOuts =
                configurationContext.getAxisConfiguration().getTransportsOut();

            for (TransportInDescription trpIn : trpIns.values()) {
                TransportListener trpLst = trpIn.getReceiver();

                if (trpLst instanceof ManagementSupport) {
                    int inUse = ((ManagementSupport) trpLst).getActiveThreadCount();
                    int inQue = ((ManagementSupport) trpLst).getQueueSize();

                    if (inUse+inQue > 0) {
                        log.info("Transport Listsner : " + trpIn.getName() +
                            " currently using : " + inUse + " threads with " +
                            inQue + " requests already queued..");

                        pendingThreads = (inUse + inQue);
                    }
                }
            }

            for (TransportOutDescription trpOut : trpOuts.values()) {
                TransportSender trpSnd = trpOut.getSender();

                if (trpSnd instanceof ManagementSupport) {
                    int inUse = ((ManagementSupport) trpSnd).getActiveThreadCount();
                    int inQue = ((ManagementSupport) trpSnd).getQueueSize();

                    if (inUse+inQue > 0) {
                        log.info("Transport Sender : " + trpOut.getName() +
                            " currently using : " + inUse + " threads with " +
                            inQue + " requests already queued..");

                        pendingThreads += (inUse + inQue);
                    }
                }
            }

            int pendingCallbacks = ServerManager.getInstance().pendingCallbacks();
            if (pendingCallbacks > 0) {
                log.info("Waiting for : " + pendingCallbacks + " callbacks/replies..");
            }

            int runningTasks = 0;
            try {
                if (getTaskScheduler() != null) {
                    runningTasks = getTaskScheduler().
                        getCurrentlyExecutingJobs().size();
                    if (runningTasks > 0) {
                        log.info("Waiting for : " + runningTasks + " tasks to complete..");
                    }
                }
            } catch (SchedulerException e) {
                log.error("Error waiting for scheduled tasks to complete", e);
            }


            // safe to shutdown is when used listner threads and callbacks are all zero
            safeToShutdown = (pendingThreads + pendingCallbacks + runningTasks == 0);

            if (safeToShutdown) {
                log.info("All transport threads and tasks are idle and no pending callbacks..");

            } else {
                if (System.currentTimeMillis() > endTime) {
                    log.info("The instance could not be gracefully shutdown in : " +
                        (waitMillis/1000) + " seconds. Shutdown immidiate..");
                    safeToShutdown = true;
                    stop();

                } else {
                    log.info("Waiting for a maximum of another " +
                        (endTime - System.currentTimeMillis())/1000 +
                        " seconds until transport threads and tasks become idle, and callbacks complete..");
                    try {
                        Thread.sleep(2000);
                    } catch (InterruptedException ignore) {}
                }
            }
        }
        log.info("Graceful shutdown request completed in " +
            (System.currentTimeMillis() - (endTime-waitMillis))/1000 + " seconds");
    }

    /**
     * Put transport listsners and senders into maintenence mode
     */
    public void startMaintenance() {
        log.info("Putting transport listners, senders and tasks into maintenence mode..");

        // put listers into maintenence
        Map<String, TransportInDescription> trpIns =
            configurationContext.getAxisConfiguration().getTransportsIn();

        for (TransportInDescription trpIn : trpIns.values()) {
            TransportListener trpLst = trpIn.getReceiver();
            if (trpLst instanceof ManagementSupport) {
                try {
                    ((ManagementSupport) trpLst).pause();
                } catch (AxisFault axisFault) {
                    log.error("Error putting transport listener for : "
                        + trpIn.getName() + " into maintenence");
                }
            }
        }

        // put senders into maintenence
        Map<String, TransportOutDescription> trpOuts =
            configurationContext.getAxisConfiguration().getTransportsOut();
        for (TransportOutDescription trpOut : trpOuts.values()) {
            TransportSender trpSnd = trpOut.getSender();
            if (trpSnd instanceof ManagementSupport) {
                try {
                    ((ManagementSupport) trpSnd).pause();
                } catch (AxisFault axisFault) {
                    log.error("Error putting transport sender for : "
                        + trpOut.getName() + " into maintenence");
                }
            }
        }

        // put tasks on hold
        try {
            if (getTaskScheduler() != null) {
                getTaskScheduler().pauseAll();
            }
        } catch (SchedulerException e) {
            log.error("Error pausing scheduled tasks", e);
        }

        log.info("[ESB] Entered maintenence mode");
    }

    /**
     * Resume normal operations for transports already entered into maintenence mode
     */
    public void endMaintenance() {
        log.info("Resuming transport listners and senders from maintenence mode..");

        // put listers into maintenence
        Map<String, TransportInDescription> trpIns =
            configurationContext.getAxisConfiguration().getTransportsIn();

        for (TransportInDescription trpIn : trpIns.values()) {
            TransportListener trpLst = trpIn.getReceiver();
            if (trpLst instanceof ManagementSupport) {
                try {
                    ((ManagementSupport) trpLst).resume();
                } catch (AxisFault axisFault) {
                    log.error("Error putting transport listener for : "
                        + trpIn.getName() + " into maintenence");
                }
            }
        }

        // put senders into maintenence
        Map<String, TransportOutDescription> trpOuts =
            configurationContext.getAxisConfiguration().getTransportsOut();
        for (TransportOutDescription trpOut : trpOuts.values()) {
            TransportSender trpSnd = trpOut.getSender();
            if (trpSnd instanceof ManagementSupport) {
                try {
                    ((ManagementSupport) trpSnd).resume();
                } catch (AxisFault axisFault) {
                    log.error("Error putting transport sender for : "
                        + trpOut.getName() + " into maintenence");
                }
            }
        }

        // Resume tasks
        try {
            if (getTaskScheduler() != null) {
                getTaskScheduler().resumeAll();
            }
        } catch (SchedulerException e) {
            log.error("Error pausing scheduled tasks", e);
        }


        log.info("[ESB] Resumed normal operations from maintenence mode");
    }

    /**
     * This implementation does not validate the user against a database. In future it
     * is envisioned that the ESB will reuse the common authentication components shared
     * between WSAS and ESB
     * @param username the username
     * @param password password
     * @return true if the user is authenticated successfully
     */
    public boolean authenticate(String username, String password) {

        OMElement docElement = sbc.getDocumentElement();
        if (docElement == null) {
            handleException("ESB configuration (server.xml) document element is null");
        }

        // get all user elements /ESBUsers/User
        List nodeList = null;
        try {
            SimpleNamespaceContext nsCtx = new SimpleNamespaceContext();
            nsCtx.addNamespace("esb", "http://www.wso2.org/projects/esb");
            XPath xp = new AXIOMXPath("//esb:ESBUsers/esb:User");
            xp.setNamespaceContext(nsCtx);
            nodeList = xp.selectNodes(docElement);

        } catch (JaxenException e) {
            handleException("Error creating XPath : //esb:ESBUsers/esb:User", e);
        }

        if (nodeList == null) {
            return false;
        }

        // for each user element get username element
        // 		if username == givenUsername then
        //			get password element
        // 			if password == givenPassword then
        //				return true
        // goto next
        // if no username matches then
        //		return false
        for (Iterator iter = nodeList.iterator(); iter.hasNext();) {

            OMElement userElement = (OMElement) iter.next();
            OMElement usernameElement = userElement.getFirstChildWithName(USERNAME_QN);

            if (usernameElement == null) {
                handleException("Required element, 'ESBUsers/User/Username' not found!");
            }

            if (username != null && usernameElement != null
                && username.equals(usernameElement.getText().trim())) {

                OMElement passwordElement = userElement.getFirstChildWithName(PASSWORD_QN);
                if (passwordElement == null) {
                    handleException("Required element, 'ESBUsers/User/Password' not found!");
                }

                if (password != null && passwordElement != null
                    && password.equals(passwordElement.getText().trim())) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Starts the Derby database instance in network mode for multi user access
     * Waits till the server can be pinged, upto a maximum of 10 seconds
     *
     * @throws ServiceBusException
     */
    private void startDataBaseServer() throws ServiceBusException {

        try {
            networkServerControl = new NetworkServerControl(
                InetAddress.getByName("localhost"), sbc.getDerbyPort());
            networkServerControl.start(null);

            for (int i = 0; i < 10; i++) {
                Thread.sleep(1000);
                try {
                    networkServerControl.ping();
                    Properties properties = networkServerControl.getCurrentProperties();
                    String host = properties.getProperty("derby.drda.host");
                    int port = Integer.parseInt(properties.getProperty("derby.drda.portNumber"));
                    log.info("Database server started on " + host + " over port " + port);
                    return;
                } catch (Exception ignore) {}
            }
            handleFatal("Derby database server failed to start within 10 Seconds");

        } catch (Exception e) {
            handleException("Error starting the embedded Derby database", e);
        }
    }

    /**
     * Requests the Derby DB to stop itself
     *
     * @throws ServiceBusException
     */
    private void stopDataBaseServer() throws ServiceBusException {
        if (networkServerControl != null) {
            try {
                networkServerControl.shutdown();
                networkServerControl = null;
                log.info("Embedded Derby database server shutdown");
            } catch (Exception e) {
                handleException("Error stopping the embedded Derby database", e);
            }
        }
    }

    /**
     * Assign deafult values for synapse configuration which are specific to ESB.
     *
     * @param cfgCtx ConfigurationContext
     */
    private void setESBDefaults(ConfigurationContext cfgCtx) {

        Parameter configParameter = cfgCtx.
                getAxisConfiguration().getParameter(SynapseConstants.SYNAPSE_CONFIG);

        Object o = configParameter.getValue();
        if (o != null && o instanceof SynapseConfiguration) {

            SynapseConfiguration synConfig = (SynapseConfiguration) o;

            // set default registry to ESB registry
            if (synConfig.getRegistry() == null) {
                Registry registry = new ESBRegistry();
                Properties props = new Properties();
                props.put("root", "file:registry");
                props.put("cachableDuration", "15000");
                registry.init(props);
                synConfig.setRegistry(registry);
            }
        }
    }

    /**
     * Generate the UI pages and cache it in the ConfigurationContext
     * @throws ServiceBusException on error
     */
    private void generateUIPages() throws ServiceBusException {

        if (configurationContext != null &&
            configurationContext.getProperty(ServiceBusConstants.GENERATED_PAGES) != null) {
            return;
        }

        Hashtable fileContents = new Hashtable();
        try {
            UIProcessor.createPages(
                sbc.getEsbHome() + File.separator + "webapp" + File.separator,
                ServiceBusConstants.UI_EXTENSIONS_CONFIG_XML, fileContents);
        } catch (UIProcessingException e) {
            handleException("Error generating page UIs", e);
        }
        configurationContext.setProperty(ServiceBusConstants.GENERATED_PAGES, fileContents);
    }

    /**
     * Start the statistics processing thread
     */
    private void startStatisticsReporter() {

        Parameter synapseEnvParameter =
            configurationContext.getAxisConfiguration().getParameter(SynapseConstants.SYNAPSE_ENV);

        if (synapseEnvParameter != null) {
            try {
                int delaySecs = sbc.getFirstPropertyAsInt("StatisticsInterval", 60);
                SynapseEnvironment synEnv = (SynapseEnvironment) synapseEnvParameter.getValue();
                statisticsReporterThread = new StatisticsReporterThread(synEnv);
                statisticsReporterThread.setDelay(delaySecs * 1000);
                statisticsReporterThread.start();
            } catch (Exception e) {
                handleException("Error starting the statistics processing thread", e);
            }
        }
    }

    /**
     * Stop the statistics processing thread
     */
    private void stopStatisticsReporter() {
        if (statisticsReporterThread != null && statisticsReporterThread.isAlive()) {
            statisticsReporterThread.shutdown();
            statisticsReporterThread.interrupt();
        }
    }

    /**
     * Start an RMI Registry on the specified port and initialize JMX control
     */
    private void startJMXService() {

        log.info("Starting the JMX Connector");

        // start an RMI Registry using the specified port.. or next available
        int rmiRegPort = sbc.getFirstPropertyAsInt("RMIRegistryPort", 1099);
        while (true) {
            try {
                if (rmiRegistry == null) {
                    rmiRegistry = LocateRegistry.createRegistry(rmiRegPort);
                }
                break;
            } catch (Exception ignored) {
                log.warn("Couldn't create a RMI registry using port : " + rmiRegPort +
                        ". Trying alternate ports..");
                rmiRegPort++;
            }
        }
        sbc.setRmiPort(rmiRegPort);

        // Register ServiceBusControlMBean
        MBeanServer mbs = ManagementFactory.getPlatformMBeanServer();

        try {
            JMXServiceURL jmxServiceURL = new JMXServiceURL(
                JMX_PREFIX + InetAddress.getLocalHost().getHostAddress() + ":" + rmiRegPort + "/esb");

            // Security credentials are included in the env Map
            HashMap<String, JMXAuthenticator> env = new HashMap<String, JMXAuthenticator>();
            env.put(JMXConnectorServer.AUTHENTICATOR, new ESBJMXAuthenticator());

            jmxConnectorServer =
                    JMXConnectorServerFactory.newJMXConnectorServer(jmxServiceURL, env, mbs);
            jmxConnectorServer.start();

            log.info("JMX consoles can connect to : " + jmxServiceURL);
        } catch (Exception e) {
            handleException("Couldn't initialize the JMX connector using port : " + rmiRegPort, e);
        }
    }

    /**
     * Stop the JMX connector and hutdown the RMI registry
     */
    private void stopJMXService() {
        if (jmxConnectorServer != null && jmxConnectorServer.isActive()) {
            try {
                log.info("Shutting down the JMX connector ");
                jmxConnectorServer.stop();
            } catch (IOException e) {
                handleException("Error Shutting down the JMX Connector", e);
            }
        }

        if (jmxConnectorServer == null || !jmxConnectorServer.isActive()) {
            if (rmiRegistry != null) {
                try {
                    UnicastRemoteObject.unexportObject(rmiRegistry, true);
                } catch (NoSuchObjectException e) {
                    handleException("Error stoping local RMI registry", e);
                }
            }
        }
    }

    /**
     * Register the ESB Management MBeans
     */
    private void registerMBeans() {
        MBeanServer mbs = ManagementFactory.getPlatformMBeanServer();
        String jmxAgentName = sbc.getJmxAgentName();

        registerMBean(mbs, new ServerStatusView(),
                jmxAgentName + ":Type=Admin,ConnectorName=ServerStatus");
        registerMBean(mbs, new ServerStatisticsView(),
                jmxAgentName + ":Type=Statistics,ConnectorName=ServerStatistics");
        registerMBean(mbs, new SequenceStatisticsView(),
                jmxAgentName + ":Type=Statistics,ConnectorName=SequenceStatistics");
        registerMBean(mbs, new ProxyServiceStatisticsView(),
                jmxAgentName + ":Type=Statistics,ConnectorName=ProxyServiceStatistics");
        registerMBean(mbs, new EndpointStatisticsView(),
                jmxAgentName + ":Type=Statistics,ConnectorName=EndpointStatistics");
        registerMBean(mbs, new SynapseServiceStatisticsView(),
                jmxAgentName + ":Type=Statistics,ConnectorName=SynapseServiceStatistics");
    }

    /**
     * Unregister the ESB Management MBeans when ServiceBusManager is stopped
     */
    private void unregisterMBeans() {

        String jmxAgentName = sbc.getJmxAgentName();
        unregisterMBean(
            jmxAgentName + ":Type=Admin,ConnectorName=ServerStatus");
        unregisterMBean(
            jmxAgentName + ":Type=Statistics,ConnectorName=ServerStatistics");
        unregisterMBean(
            jmxAgentName + ":Type=Statistics,ConnectorName=SequenceStatistics");
        unregisterMBean(
            jmxAgentName + ":Type=Statistics,ConnectorName=ProxyServiceStatistics");
        unregisterMBean(
            jmxAgentName + ":Type=Statistics,ConnectorName=EndpointStatistics");
        unregisterMBean(
            jmxAgentName + ":Type=Statistics,ConnectorName=SynapseServiceStatistics");
    }

    //---------------------- utility methods ---------------------------
    /**
     * Register a given MBean instance with the supplied name on the given MBeanServer
     * @param mbs
     * @param mbeanInstance
     * @param objectName
     */
    private void registerMBean(MBeanServer mbs, Object mbeanInstance, String objectName) {
        try {
            ObjectName name = new ObjectName(objectName);
            Set set = mbs.queryNames(name, null);
            if (set != null && set.isEmpty()) {
                mbs.registerMBean(mbeanInstance, name);
            } else {
                mbs.unregisterMBean(name);
                mbs.registerMBean(mbeanInstance, name);
            }
        } catch (Exception e) {
            log.warn("Error registering MBean with objectname ' " + objectName +
                    " ' for JMX management", e);
        }
    }

    /**
     * Unregister an MBean
     * @param objectName name as a string
     */
    private void unregisterMBean(String objectName) {
        try {
            ObjectName name = new ObjectName(objectName);
            ManagementFactory.getPlatformMBeanServer().unregisterMBean(name);
        } catch (Exception ignore) {}
    }

    private static void handleFatal(String msg) throws ServiceBusException {
        log.fatal(msg);
        throw new ServiceBusException(msg);
    }

    private static void handleException(String msg, Exception e) throws ServiceBusException {
        log.error(msg, e);
        throw new ServiceBusException(msg, e);
    }

    private static void handleException(String msg) {
        log.error(msg);
        throw new ServiceBusException(msg);
    }

    private Scheduler getTaskScheduler() throws SchedulerException {
        try {
            return DirectSchedulerFactory.getInstance().getScheduler();
        } catch (SchedulerException e) {
            if (e.getMessage().indexOf("must call createRemoteScheduler") == -1) {
                throw e;
            }
        }
        return null;
    }

    //---------------------- public getters and setters ---------------------------

    public ConfigurationContext getConfigurationContext() {
        return configurationContext;
    }

    public ClassLoader getClassLoader() {
        return classLoader;
    }
}
