/*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */

package org.wso2.esb.registry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.synapse.SynapseException;
import org.apache.synapse.registry.Registry;
import org.apache.synapse.registry.url.SimpleURLRegistry;
import org.wso2.esb.ServiceBusConstants;
import org.wso2.esb.ServiceBusException;
import org.wso2.esb.ServiceBusConfiguration;
import org.wso2.registry.RegistryException;
import org.wso2.registry.app.RemoteRegistry;
import org.wso2.registry.jdbc.EmbeddedRegistry;
import org.wso2.registry.jdbc.InMemoryEmbeddedRegistry;
import org.wso2.registry.jdbc.utils.RegistryDataSource;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * Finds the any adapters that provides synapse registry API
 */
public class RegistryAdapterFactory {

    private static final Log log = LogFactory.getLog(RegistryAdapterFactory.class);

    private static final String LOCAL = "local";
    private static final String REMOTE = "remote";
    private static final String URL = "url";
    private static final String IN_MEMORY = "inMemory";
    private static final String ESB_I6 = "esb-1.6";

    /**
     * Returns the registry implementation
     *
     * @param esbConfiguration The configuration that hold registry configuration data
     * @return A registry instance
     */
    public static Registry createRegistryAdapter(ServiceBusConfiguration esbConfiguration) {

        //Gets the type of registry
        String type = esbConfiguration
                .getFirstProperty
                        (ServiceBusConstants.REGISTRY + "." +
                                ServiceBusConstants.TYPE);
        //Gets the import root
        String importRoot = esbConfiguration
                .getFirstProperty
                        (ServiceBusConstants.REGISTRY + "." +
                                ServiceBusConstants.IMPORT_ROOT);

        if (ESB_I6.equals(type)) {

            return new ESBRegistry();

        } else if (LOCAL.equals(type) || REMOTE.equals(type) || IN_MEMORY.equals(type)) {

            WSO2RegistryAdapter adapter;

            if (LOCAL.equals(type)) {
                adapter = new WSO2RegistryAdapter(createJDBCRegistry(esbConfiguration));
            } else if (REMOTE.equals(type)) {
                adapter = new WSO2RegistryAdapter(createRemoteRegistry(esbConfiguration));
            } else {
                adapter = new WSO2RegistryAdapter(createInMemoryRegistry());
            }

            if (importRoot != null && !"".equals(importRoot)) {
                File importDir = getImportRootDirectory(importRoot);
                if (importDir != null) {
                    adapter.importResources(importDir, "");
                }
            }
            return adapter;

        } else if (URL.equals(type)) {

            return new SimpleURLRegistry();

        } else {

            //Default registry will be inMemory registry
            WSO2RegistryAdapter adapter = new WSO2RegistryAdapter(createInMemoryRegistry());

            if (importRoot != null && !"".equals(importRoot)) {
                File importDir = getImportRootDirectory(importRoot);
                if (importDir != null) {
                    adapter.importResources(importDir, "");
                }
            }
            return adapter;
        }
    }

    private static File getImportRootDirectory(String path) {
        try {
            URL url = new URL(path);
            path = url.getPath();
        } catch (MalformedURLException e) {
            // don't do any thing if this is not a valid URL
        }
        File file = new File(path);
        if (!file.isAbsolute()) {
            String esbHome = ServiceBusConfiguration.getInstance().getEsbHome();
            if (!esbHome.endsWith("/")) {
                esbHome = esbHome + "/";
            }
            file = new File(esbHome + path);
        }
        return file;
    }

    /**
     * Helper method to create a InMemory registry
     *
     * @return InMemory registry instance
     */
    private static org.wso2.registry.Registry createInMemoryRegistry() {

        try {

            return new InMemoryEmbeddedRegistry().getSystemRegistry();

        } catch (RegistryException e) {
            handleException("Error when initiating a in memory registry", e);
        }

        // never executes - just to make the compiler happy
        return null;
    }

    /**
     * Helper method to create a JDBC registry
     *
     * @param esbConfiguration ESB configuration instance to be used to get configuration data
     * @return A JDBC registry instance
     */
    private static org.wso2.registry.Registry createJDBCRegistry(
        ServiceBusConfiguration esbConfiguration) {

        String registryDriverClass = esbConfiguration
                .getFirstProperty
                        (ServiceBusConstants.REGISTRY + "." +
                                ServiceBusConstants.CONFIGURATION +
                                "." + ServiceBusConstants.DRIVER_CLASS);
        String registryUrl = esbConfiguration
                .getFirstProperty(
                        ServiceBusConstants.REGISTRY + "." +
                                ServiceBusConstants.CONFIGURATION +
                                "." + ServiceBusConstants.URL);
        String registryUsername = esbConfiguration
                .getFirstProperty(ServiceBusConstants.REGISTRY + "." +
                        ServiceBusConstants.CONFIGURATION +
                        "." + ServiceBusConstants.USERNAME);
        String registryPassword = esbConfiguration
                .getFirstProperty(ServiceBusConstants.REGISTRY + "." +
                        ServiceBusConstants.CONFIGURATION +
                        "." + ServiceBusConstants.PASSWORD);

        if (registryDriverClass == null || "".equals(registryDriverClass) ||
                registryUrl == null || "".equals(registryUrl)) {
            handleException("Cannot find DataBase connection details pertaining to " +
                    "the registry. The driverClass and the url are mandatory.");
        }

        try {

            RegistryDataSource dataSource = new RegistryDataSource(
                    registryUrl, registryDriverClass, registryUsername, registryPassword);

            EmbeddedRegistry embeddedRegistry = new EmbeddedRegistry(dataSource);
            return embeddedRegistry.getSystemRegistry();

        } catch (RegistryException e) {
            handleException("Error when initiating a JDBC registry", e);
        }
        throw new ServiceBusException("JDBC registry cannot be null");
    }

    /**
     * Helper method to initiate a Remote Registry instance
     *
     * @param esbConfiguration ESB configuration instance to be used to get configuration data
     * @return A remote registry instance
     */
    private static org.wso2.registry.Registry createRemoteRegistry(
        ServiceBusConfiguration esbConfiguration) {

        String registryUrl = esbConfiguration
                .getFirstProperty(
                        ServiceBusConstants.REGISTRY + "." +
                                ServiceBusConstants.CONFIGURATION +
                                "." + ServiceBusConstants.URL);
        String registryUsername = esbConfiguration
                .getFirstProperty(ServiceBusConstants.REGISTRY + "." +
                        ServiceBusConstants.CONFIGURATION +
                        "." + ServiceBusConstants.USERNAME);
        String registryPassword = esbConfiguration
                .getFirstProperty(ServiceBusConstants.REGISTRY + "." +
                        ServiceBusConstants.CONFIGURATION +
                        "." + ServiceBusConstants.PASSWORD);

        if ("".equals(registryUrl)) {
            handleException("The url is mandatory for remote registry.");
        }

        try {
            return new RemoteRegistry(registryUrl,
                    registryUsername, registryPassword);
        } catch (MalformedURLException e) {
            handleException("Error when initiating a remote registry instance", e);
        }
        throw new ServiceBusException("");
    }

    private static void handleException(String msg, Exception e) {
        log.error(msg, e);
        throw new SynapseException(msg, e);
    }

    private static void handleException(String msg) {
        log.error(msg);
        throw new SynapseException(msg);
    }

}
