/*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */

package org.wso2.esb.registry;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMException;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMNode;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.synapse.SynapseException;
import org.apache.synapse.config.Entry;
import org.apache.synapse.registry.AbstractRegistry;
import org.apache.synapse.registry.RegistryEntry;
import org.apache.synapse.registry.RegistryEntryImpl;
import org.apache.synapse.util.SynapseBinaryDataSource;
import org.wso2.esb.ServiceBusConstants;
import org.wso2.registry.*;
import org.wso2.registry.utils.RegistryClientUtils;

import javax.activation.DataHandler;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

/**
 * This adopts the WSO2 registry as a synapse registry.
 */
public class WSO2RegistryAdapter extends AbstractRegistry {

    private static final Log log = LogFactory.getLog(WSO2RegistryAdapter.class);

    /* Static constants to name some properties         */
    private static final String EXPIRY_TIME = "expiryTime";
    private static final String CACHABLE_DURATION = "cachableDuration";

    /* Max number of resources */
    private static final int MAX_KEYS = 200;

    /* WSO2Registy instance , All of the synapse registry API calls will be delegated to this  */
    private org.wso2.registry.Registry registry;

    /* The root path which is used as a parent path for any resource */
    private String rootPath;

    public WSO2RegistryAdapter(org.wso2.registry.Registry registry) {
        this.registry = registry;
    }

    /**
     * Perform an actual lookup for for an XML resource as an OMNode for the given key
     *
     * @param key the key for the registry lookup
     * @return the XML content from the registry as an OMNode
     */
    public OMNode lookup(String key) {

        if (key == null) {
            handleException("Resource cannot be found.");
        }

        if ("".equals(key)) {
            handleException("Resource cannot be empty");
        }

        if (log.isDebugEnabled()) {
            log.debug("==> Repository fetch of resource with key : " + key);
        }

        try {

            Resource resource = getResource(key);
            if (resource == null) {
                return null;
            }

            InputStream input = resource.getContentStream();
            if (input == null) {
                return null;
            }

            BufferedInputStream inputStream = new BufferedInputStream(input);
            OMNode result = null;
            try {
                XMLStreamReader parser = XMLInputFactory.newInstance().
                        createXMLStreamReader(inputStream);
                StAXOMBuilder builder = new StAXOMBuilder(parser);
                result = builder.getDocumentElement();

            } catch (OMException ignored) {

                if (log.isDebugEnabled()) {
                    log.debug("The resource at the provided URL isn't " +
                            "well-formed XML,So,takes it as a text");
                }

                try {
                    inputStream.close();
                } catch (IOException e) {
                    log.error("Error in closing the input stream. ", e);
                }
                resource = getResource(key);
                result = readNonXML(resource.getContentStream(), resource.getMediaType());

            } catch (XMLStreamException ignored) {

                if (log.isDebugEnabled()) {
                    log.debug("The resource at the provided URL isn't " +
                            "well-formed XML,So,takes it as a text");
                }

                try {
                    inputStream.close();
                } catch (IOException e) {
                    log.error("Error in closing the input stream. ", e);
                }
                resource = getResource(key);
                result = readNonXML(resource.getContentStream(), resource.getMediaType());

            } finally {
                try {
                    result.detach();
                    inputStream.close();
                } catch (IOException e) {
                    log.error("Error in closing the input stream.", e);
                }

            }
            return result;
        } catch (RegistryException e) {
            handleException("\"Error when fetching a resource with key : \"" + key, e);
        }
        return null;
    }

    /**
     * Helper method to handle non-XMl resources
     *
     * @param inputStream resource content as stream
     * @param contentType content-type
     * @return The content as an OMNode
     */
    private OMNode readNonXML(InputStream inputStream, String contentType) {

        try {

            OMFactory omFactory = OMAbstractFactory.getOMFactory();
            return omFactory.createOMText(
                    new DataHandler(new SynapseBinaryDataSource(inputStream,
                            contentType)), true);

        } catch (IOException e) {
            handleException("Error when getting a stream from resource's content ", e);
        }
        return null;
    }

    /**
     * This is the publicly used interface to the registry. It will fetch
     * the content from the registry and cache if required.
     *
     * @param entry the registry Entry
     * @return the value from the registry or local cache
     * @see AbstractRegistry
     */
    public Object getResource(Entry entry) {
        return super.getResource(entry);
    }

    /**
     * Get the registry entry for the given key
     *
     * @param entryKey the registry key
     * @return The registry entry for the given key
     */
    public RegistryEntry getRegistryEntry(String entryKey) {

        RegistryEntryImpl entry = new RegistryEntryImpl();
        Resource resource = getResource(entryKey);

        if (resource != null) {

            entry.setKey(getResourcePath(entryKey));
            entry.setName(resource.getPath());

            if (resource instanceof Collection) {
                entry.setType(ServiceBusConstants.folder);
            } else {
                entry.setType(ServiceBusConstants.file);
            }

            entry.setDescription("Resource at : " + resource.getPath());
            entry.setLastModified(resource.getLastModified().getTime());
            entry.setVersion(resource.getLastModified().getTime());

            String cachableDuration = resource.getProperty(CACHABLE_DURATION);
            long cacheTime = 0;
            if (cachableDuration != null) {
                try {
                    cacheTime = Long.parseLong(cachableDuration);
                } catch (NumberFormatException e) {
                    handleException("Couldn't pass the cachableDuration as a long", e);
                }
                entry.setCachableDuration(cacheTime);
            }
        }

        return entry;
    }

    /**
     * Returns the child elements of a given registry entry
     *
     * @param entry - parent registry entry
     * @return Array of child registry entries of the given parent registry entry
     */
    public RegistryEntry[] getChildren(RegistryEntry entry) {

        if (entry == null) {
            // give the children of the root
            // null or key = "" stands for root

            RegistryEntryImpl registryEntry = new RegistryEntryImpl();
            registryEntry.setKey(removePathSeparator(rootPath));
            entry = registryEntry;
        } else {
            if ("".equals(entry.getKey())) {
                ((RegistryEntryImpl) entry).setKey(removePathSeparator(rootPath));
            }
        }

        Resource resource = getResource(entry.getKey());
        if (resource == null) {
            return new RegistryEntry[]{new RegistryEntryImpl()};
        }

        if (resource instanceof Collection) {

            CollectionImpl collection = (CollectionImpl) resource;
            String[] children = collection.getChildren();
            List<RegistryEntry> entryList = new ArrayList<RegistryEntry>();

            for (String child : children) {

                String key = removePathSeparator(child);

                Resource childResource = getResource(key);
                if (childResource == null) {
                    continue;
                }

                RegistryEntryImpl registryEntryEmbedded = new RegistryEntryImpl();
                registryEntryEmbedded.setKey(getResourcePath(key));

                if (childResource instanceof Collection) {
                    registryEntryEmbedded.setType(ServiceBusConstants.folder);
                } else {
                    registryEntryEmbedded.setType(ServiceBusConstants.file);
                }

                entryList.add(registryEntryEmbedded);
            }

            return entryList.toArray(new RegistryEntry[entryList.size()]);
        } else {
            ((RegistryEntryImpl) entry).setType(ServiceBusConstants.file);
            return new RegistryEntry[]{entry};
        }

    }

    /**
     * Returns all descendant entries of the given registry entry
     *
     * @param entry - parent registry entry
     * @return Array of decendant registry entries of the given registry entry
     */
    public RegistryEntry[] getDescendants(RegistryEntry entry) {

        ArrayList<RegistryEntry> list = new ArrayList<RegistryEntry>();
        RegistryEntry[] entries = getChildren(entry);
        
        if (entries != null) {
            for (RegistryEntry entry1 : entries) {
                if (list.size() > MAX_KEYS) {
                    break;
                }
                fillDescendants(entry1, list);
            }
        }

        RegistryEntry[] descendants = new RegistryEntry[list.size()];
        for (int i = 0; i < list.size(); i++) {
            descendants[i] = list.get(i);
        }

        return descendants;
    }

    /**
     * Return the name of the implementation class
     *
     * @return name of the registry provider implementation class name
     */
    public String getProviderClass() {
        return super.getProviderClass();
    }

    /**
     * Return the list of configuration properties set on this instance
     *
     * @return a Map of configuration properties
     */
    public Properties getConfigurationProperties() {

        return super.getConfigurationProperties();
    }

    /**
     * Deletes a resource in a given path
     *
     * @param path The resource path to be deleted
     */
    public void delete(String path) {
        try {
            String resolvedPath = calculatePath(rootPath, path);
            if (registry.resourceExists(resolvedPath)) {
                registry.delete(resolvedPath);
            }
        } catch (RegistryException e) {
            handleException("Error when deleting a resource at path :" + path, e);
        }
    }

    /**
     * Creates a new resource in a given path
     *
     * @param resourcePath The path to be used as resource path
     * @param isDirectory  This is resource is a collection
     */
    public void newResource(String resourcePath, boolean isDirectory) {

        // Gets child name
        String childName = getResourceName(resourcePath);
        //Gets parent name
        String parentPath = getParentPath(resourcePath);
        //Gets parent resource
        Resource parent = getResource(parentPath);
        parentPath = parent.getPath();

        //Corrected path
        String path = calculatePath(parentPath, childName);

        //Getting cache time and expiration time from parent ,
        //as child should inherit parents properties
        long cacheTime = getCachableDuration(); // roots cache time
        long expirationTime;

        if (parent != null) {

            String cachableDuration = parent.getProperty(CACHABLE_DURATION);
            if (cachableDuration != null) {
                try {
                    cacheTime = Long.parseLong(cachableDuration);
                } catch (NumberFormatException e) {
                    handleException("Couldn't pass the cachableDuration as a long", e);                    
                }
            }
        }

        //new expiration time
        expirationTime = System.currentTimeMillis() + cacheTime;

        if (isDirectory) {

            CollectionImpl collection = new CollectionImpl();
            collection.setPath(path);
            collection.setProperty(CACHABLE_DURATION, String.valueOf(cacheTime));
            collection.setProperty(EXPIRY_TIME, String.valueOf(expirationTime));
            try {
                registry.put(path, collection);
            } catch (RegistryException e) {
                handleException("Error when setting a resource in the path : " + path, e);
            }

        } else {

            ResourceImpl resource = new ResourceImpl();
            resource.setPath(path);
            resource.setProperty(CACHABLE_DURATION, String.valueOf(cacheTime));
            resource.setProperty(EXPIRY_TIME, String.valueOf(expirationTime));
            try {
                registry.put(path, resource);
            } catch (RegistryException e) {
                handleException("Error when setting a resource in the path : " + path, e);
            }
        }
    }

    /**
     * Updates the content of a resource in the given path with given content
     *
     * @param path  The resource path
     * @param value The resource content to be set
     */
    public void updateResource(String path, Object value) {

        Resource resource = getResource(path);
        if (resource != null) {

            if (value instanceof OMNode) {
                try {
                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                    ((OMNode) value).serialize(baos);
                    resource.setContent(baos.toByteArray());
                } catch (XMLStreamException e) {
                    handleException("Error when serializing OMNode " + value, e);
                }

            } else {
                resource.setContent(value);
            }

            try {
                registry.put(resource.getPath(), resource);
            } catch (RegistryException e) {
                handleException("Error when setting a resource in the path : " + path, e);
            }
        }

    }

    /**
     * Updates the resource meta-data
     *
     * @param entry The resource meta-data
     */
    public void updateRegistryEntry(RegistryEntry entry) {

        String key = entry.getKey();
        Resource resource = getResource(key);

        if (resource != null) {
            resource.setProperty(CACHABLE_DURATION,
                    String.valueOf(entry.getCachableDuration()));
            resource.setProperty(EXPIRY_TIME,
                    String.valueOf(System.currentTimeMillis() + entry.getCachableDuration()));
            try {
                registry.put(resource.getPath(), resource);
            } catch (RegistryException e) {
                handleException("Error when setting a resource in the path : " + key, e);
            }
        }
    }

    /**
     * Initializes the registry with given properties
     *
     * @param properties The configuration properties
     */
    public void init(Properties properties) {

        super.init(properties);

        String root = properties.getProperty("root");
        if (root == null || "".equals(root)) {
            rootPath = "/";
            return;
        }
        try {
            URL url = new URL(root);
            root = url.getPath();
            if (url.getProtocol().equals("file")) {
                if (!root.endsWith("/")) {
                    root = root + "/";
                }
                if (!root.startsWith("/")) {
                    root = "/" + root;
                }

            }
        } catch (MalformedURLException e) {
            // don't do any thing if this is not a valid URL
        }
        rootPath = root;
    }

    /**
     * Imports all the resources under a given directory to the registry
     *
     * @param file   The root directory
     * @param toPath The root path for registry
     */
    public void importResources(File file, String toPath) {

        if (file == null) {
            return;
        }
        if (file.isDirectory()) {
            String[] children = file.list();
            for (String strChild : children) {

                try {
                    File child = new File(file, strChild);
                    String path = removePathSeparator(toPath) + "/" + child.getName();

                    if (!registry.resourceExists(path)) {
                        RegistryClientUtils.importToRegistry(child,
                            toPath, registry);
                    }

                } catch (RegistryException e) {
                    handleException("Error when importing resources in " +
                        "the path : " + file.getPath(), e);
                }
            }
        }
    }

    private void fillDescendants(RegistryEntry parent, ArrayList<RegistryEntry> list) {

        RegistryEntry[] entries = getChildren(parent);
        if (entries != null) {
            for (RegistryEntry entry : entries) {
                if (list.size() > MAX_KEYS) {
                    break;
                }
                fillDescendants(entry, list);
            }
        } else {
            list.add(parent);
        }
    }


    private Resource getResource(String s) {
        Resource resource = null;
        try {
            String resolvedPath = calculatePath(rootPath, s);
            if (registry.resourceExists(resolvedPath)) {
                resource = registry.get(resolvedPath);
            }
        } catch (RegistryException e) {
            handleException("Error when fetching resource wit key " + s, e);
        }
        return resource;
    }

    private String removePathSeparator(String path) {

        if (path.endsWith(RegistryConstants.PATH_SEPARATOR)) {
            return path.substring(0, path.lastIndexOf(RegistryConstants.PATH_SEPARATOR));
        } else {
            return path;
        }
    }

    private String calculatePath(String parentPath, String resourceName) {
        String resourcePath;
        resourceName = removePathSeparator(resourceName);
        if (resourceName.startsWith(RegistryConstants.PATH_SEPARATOR)) {
            resourceName = resourceName.substring(1, resourceName.length());
        }
        if (!parentPath.startsWith(RegistryConstants.PATH_SEPARATOR)) {
            parentPath = RegistryConstants.PATH_SEPARATOR + parentPath;
        }
        if (parentPath.endsWith(RegistryConstants.PATH_SEPARATOR)) {
            resourcePath = parentPath + resourceName;
        } else {
            resourcePath = parentPath + RegistryConstants.PATH_SEPARATOR + resourceName;
        }
        return resourcePath;
    }

    private String getParentPath(String resourcePath) {

        String parentPath;
        if (resourcePath.equals(RegistryConstants.ROOT_PATH)) {
            parentPath = null;
        } else {
            if (resourcePath.lastIndexOf(RegistryConstants.PATH_SEPARATOR) == 0) {
                parentPath = RegistryConstants.ROOT_PATH;
            } else {
                parentPath = resourcePath.substring(
                        0, resourcePath.lastIndexOf(RegistryConstants.PATH_SEPARATOR));
            }
        }
        return parentPath;
    }

    private long getCachableDuration() {
        String cachableDuration = (String) properties.get("cachableDuration");
        return cachableDuration == null ? 0 : Long.parseLong(cachableDuration);
    }

    private String getResourcePath(String path) {
        path = calculatePath(RegistryConstants.PATH_SEPARATOR, path);
        if (path.startsWith(rootPath)) {
            path = path.substring(rootPath.length(), path.length());
        } else {
            String root = removePathSeparator(rootPath);
            if (root.equals(path)) {
                path = "";
            }
        }
        return path;
    }

    private String getResourceName(String path) {
        if (path != null) {
            String correctedPath = path;
            if (path.endsWith(RegistryConstants.PATH_SEPARATOR)) {
                correctedPath = path.substring(0,
                        path.lastIndexOf(RegistryConstants.PATH_SEPARATOR));
            }
            return correctedPath.substring(
                    correctedPath.lastIndexOf(RegistryConstants.PATH_SEPARATOR) + 1,
                    correctedPath.length());
        }
        return "";

    }

    private void handleException(String msg, Exception e) {
        log.error(msg, e);
        throw new SynapseException(msg, e);
    }

    private void handleException(String msg) {
        log.error(msg);
        throw new SynapseException(msg);
    }

}
