/*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */

package org.wso2.esb.services;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.axiom.om.impl.llom.factory.OMXMLBuilderFactory;
import org.apache.axis2.AxisFault;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.synapse.SynapseConstants;
import org.apache.synapse.SynapseException;
import org.apache.synapse.config.SynapseConfiguration;
import org.apache.synapse.config.xml.XMLConfigurationBuilder;
import org.apache.synapse.config.xml.XMLConfigurationSerializer;
import org.apache.synapse.core.axis2.Axis2SynapseEnvironment;
import org.apache.synapse.core.axis2.MessageContextCreatorForAxis2;
import org.apache.synapse.core.axis2.ProxyService;
import org.wso2.esb.util.XMLPrettyPrinter;

import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLInputFactory;
import java.io.*;
import java.util.Iterator;

/**
 * This is a POJO for the configuration admin service
 */
public class ConfigAdmin extends AbstractESBAdmin {

    private static final Log log = LogFactory.getLog(ConfigAdmin.class);

    public String getConfiguration() throws AxisFault {
        try {
            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            XMLConfigurationSerializer.serializeConfiguration(getSynapseConfiguration(), stream);
            XMLInputFactory factory = XMLInputFactory.newInstance();
            factory.setProperty("http://java.sun.com/xml/stream/properties/report-cdata-event", Boolean.TRUE);
            ByteArrayInputStream byteArrayInputStream = new ByteArrayInputStream(stream.toByteArray());
            StAXOMBuilder builder = OMXMLBuilderFactory.createStAXOMBuilder(
                    OMAbstractFactory.getOMFactory(),factory.createXMLStreamReader(
                    byteArrayInputStream));
            stream.reset();
            XMLPrettyPrinter.prettify(builder.getDocumentElement(), stream);
            return new String(stream.toByteArray());
        } catch (XMLStreamException e) {
            log.error("Error serializing the Synapse configuration");
            throw AxisFault.makeFault(e);
        } catch (Exception e) {
            log.error("Error serializing the Synapse configuration");
            throw AxisFault.makeFault(e);
        }
    }

    public void updateConfiguration(OMElement configElement) throws AxisFault {
        try {
            log.info("Stopping Proxy services...");
            AxisConfiguration axisCfg = getAxisConfig();
            Iterator iter = getSynapseConfiguration().getProxyServices().iterator();
            while (iter.hasNext()) {
                ProxyService proxy = (ProxyService) iter.next();
                try {
                    axisCfg.removeService(proxy.getName());
                } catch (AxisFault e) {
                    log.error("Error removing Proxy service : " + proxy.getName(), e);
                    throw e;
                }
            }

            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            configElement.getFirstElement().serialize(stream);
            SynapseConfiguration synCfg = XMLConfigurationBuilder.getConfiguration(
                new ByteArrayInputStream(stream.toByteArray()));
            synCfg.setPathToConfigFile(getSynapseConfiguration().getPathToConfigFile());
            synCfg.setAxisConfiguration(axisCfg);
            Parameter synapseCtxParam = new Parameter(SynapseConstants.SYNAPSE_CONFIG, null);
            synapseCtxParam.setValue(synCfg);
            MessageContextCreatorForAxis2.setSynConfig(synCfg);

            //set up synapse env
            ConfigurationContext configCtx = getConfigContext();
            Parameter synapseEnvParam = new Parameter(SynapseConstants.SYNAPSE_ENV, null);
            Axis2SynapseEnvironment synEnv = new Axis2SynapseEnvironment(configCtx, synCfg);
            synapseEnvParam.setValue(synEnv);
            MessageContextCreatorForAxis2.setSynEnv(synEnv);

            try {
                axisCfg.addParameter(synapseCtxParam);
                axisCfg.addParameter(synapseEnvParam);

            } catch (AxisFault e) {
                String msg =
                    "Could not set parameters '" + SynapseConstants.SYNAPSE_CONFIG +
                        "' and/or '" + SynapseConstants.SYNAPSE_ENV +
                        "'to the Axis2 configuration : " + e.getMessage();
                handleFault(log,msg,e);
            }
            // redeploy proxy services
            log.info("Re-deploying Proxy services...");
            iter = synCfg.getProxyServices().iterator();
            while (iter.hasNext()) {
                ProxyService proxy = (ProxyService) iter.next();
                proxy.buildAxisService(synCfg, axisCfg);
                log.info("Deployed Proxy service : " + proxy.getName());
                if (!proxy.isStartOnLoad()) {
                    proxy.stop(synCfg);
                }
            }
            // re-init the synapse configuration
            synCfg.init(synEnv);

            log.info("Configuration updated successfully...");

        } catch (SynapseException se) {
            handleFault(log, "Unable to update the Synapse configuration." +
                se.getMessage() + " Check log for more details", se);
        } catch (XMLStreamException e) {
            handleFault(log, "Unable to update the Synapse configuration." +
                e.getMessage() + " Check log for more details", e);
        }
    }

    public boolean saveConfigurationToDisk() throws AxisFault {
        log.info("Saving configuration..");

        SynapseConfiguration config = getSynapseConfiguration();
        try {
            FileOutputStream fos = new FileOutputStream(config.getPathToConfigFile());
            XMLConfigurationSerializer.serializeConfiguration(config, fos);
            try {
                fos.close();
                XMLPrettyPrinter.prettify(new File(config.getPathToConfigFile()));
            }
            catch (IOException e) {
                // ignore prettyfy errors
            }
            log.info("Configuration saved to disk");
            return true;
        }
        catch (XMLStreamException e) {
            handleFault(log, "Could not save changes to disk." +
                e.getMessage() + " Check log for more details", e);
        }
        catch (FileNotFoundException e) {
            handleFault(log, "Could not locate the Synapse configuration file to save changes", e);
        }
        return false;
    }

}
