/*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */

package org.wso2.esb.services;

import org.apache.axiom.om.*;
import org.apache.axis2.AxisFault;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.synapse.SynapseException;
import org.apache.synapse.config.Entry;
import org.apache.synapse.config.SynapseConfiguration;
import org.apache.synapse.config.xml.EntrySerializer;
import org.apache.synapse.config.xml.XMLConfigConstants;
import org.apache.synapse.endpoints.Endpoint;
import org.apache.synapse.mediators.base.SequenceMediator;
import org.wso2.esb.services.tos.EntryData;

import javax.xml.namespace.QName;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

/**
 * This is a POJO for Entry based administration service
 */
public class EntriesAdmin extends AbstractESBAdmin {

    private static final Log log = LogFactory.getLog(EntriesAdmin.class);

    public EntryData[] entryData() throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        Map gloabalEntiesMap = synapseConfiguration.getDefinedEntries();
        ArrayList globalEntryList = new ArrayList();
        for (Iterator ite = gloabalEntiesMap.entrySet().iterator();
             ite.hasNext();) {
            EntryData data = new EntryData();
            Map.Entry entry = (Map.Entry) ite.next();
            String key = (String) entry.getKey();
            if (entry.getValue() instanceof Entry) {
                Entry value = (Entry) entry.getValue();
                data.setName(key);
                switch (value.getType()) {
                    case Entry.REMOTE_ENTRY:
                        data.setType("Registry Key");
                        break;
                    case Entry.INLINE_TEXT:
                        data.setType("Inline Text");
                        break;
                    case Entry.INLINE_XML:
                        data.setType("Inline XML");
                        break;
                    case Entry.URL_SRC:
                        data.setType("Source URL");
                        break;
                    default:
                        data.setType("Unknown");
                        break;
                }
                if (value.getValue() instanceof String) {
                    String s = (String) value.getValue();
                    if (value.getType() == Entry.INLINE_TEXT &&
                        s.length() > 40) {
                        data.setValue(s.substring(0, 40) + "...");
                    } else {
                        data.setValue(s);
                    }
                } else if (value.getType() == Entry.URL_SRC) {
                    String s = value.getSrc().toString();
                    if (s.length() > 40) {
                        data.setValue(s.substring(0, 40) + "...");
                    } else {
                        data.setValue(s);
                    }
                } else if (value.getType() == Entry.REMOTE_ENTRY)
                    data.setValue(value.getKey());
                else
                    data.setValue("'XML Fragment'");
                globalEntryList.add(data);
            }
        }

        Collections.sort(globalEntryList, new Comparator() {
            public int compare(Object o1, Object o2) {
                return ((EntryData) o1).getName().compareToIgnoreCase(
                    ((EntryData) o2).getName());
            }
        });

        return (EntryData[]) globalEntryList
            .toArray(new EntryData[globalEntryList.size()]);
    }

    /**
     * Add a entry into the synapseConfiguration
     *
     * @param elem - Entry object to be added as an OMElement
     * @throws AxisFault if a Entry exists with the same name or if the
     *                   element provided is not a Entry element
     */
    public boolean addEntry(OMElement elem) throws AxisFault {
        try {
            if (elem.getQName().getLocalPart().equals(XMLConfigConstants
                .ENTRY_ELT.getLocalPart())) {

                String entryName = elem.getAttributeValue(new QName("name"));

                log.debug("Adding local entry with name : " + entryName);
                if (getSynapseConfiguration().getLocalRegistry().containsKey(entryName)) {
                    handleFault(log, "The name '" + entryName +
                        "' is already used within the configuration", null);
                } else {
                    OMAttribute name = elem.getAttribute(
                        new QName(XMLConfigConstants.NULL_NAMESPACE, "name"));
                    if (name == null) {
                        handleFault(log, "Unable to add local entry. Cannot find the name", null);

                    } else {
                        Entry entry = new Entry();
                        entry.setKey(name.getAttributeValue());
                        String value = elem.getAttributeValue(
                            new QName(XMLConfigConstants.NULL_NAMESPACE, "value"));
                        String src = elem.getAttributeValue(
                            new QName(XMLConfigConstants.NULL_NAMESPACE, "src"));
                        String key = elem.getAttributeValue(
                            new QName(XMLConfigConstants.NULL_NAMESPACE, "key"));
                        int type = Integer.parseInt(elem.getAttributeValue(
                            new QName(XMLConfigConstants.NULL_NAMESPACE, "type")));
                        OMElement content = elem.getFirstElement();

                        if (type == Entry.URL_SRC) {
                            entry.setType(Entry.URL_SRC);
                            try {
                                entry.setSrc(new URL(src));
                            } catch (MalformedURLException e) {
                                handleFault(log, "Invalid URL : '" + src + "'", e);
                            }
                        } else if (type == Entry.REMOTE_ENTRY) {
                            entry.setType(Entry.REMOTE_ENTRY);
                            entry.setKey(key);
                        } else if (type == Entry.INLINE_XML) {
                            entry.setType(Entry.INLINE_XML);
                            entry.setValue(content);
                        } else if (type == Entry.INLINE_TEXT) {
                            entry.setType(Entry.INLINE_TEXT);
                            entry.setValue(elem.getText());
                        } else {
                            handleFault(log, "Unable to add local entry. Invalid definition", null);
                        }
                        getSynapseConfiguration().addEntry(entryName, entry);
                    }
                }
                log.info("Local registry entry : " + entryName + " added to the configuration");
                return true;
            } else {
                handleFault(log, "Error adding local entry. Invalid definition", null);
            }
        } catch (SynapseException syne) {
            handleFault(log, "Unable to add local entry ", syne);
        }
        return false;
    }

    /**
     * Saves the entry described with the OMElement representing the entry
     *
     * @param elem - OMElement representing the entry
     * @throws AxisFault if the entry name already exists or if the Element
     *                   doesnt represent a entry element
     */
    public boolean saveEntry(OMElement elem) throws AxisFault {

        Entry entry = new Entry();
        if (elem == null) {
            handleFault(log, "Unable to save local entry. Null definition", null);
        }
        OMAttribute name = elem.getAttribute(
            new QName(XMLConfigConstants.NULL_NAMESPACE, "name"));
        String nameValue = null;
        if (name != null) {
            if (name.getAttributeValue() != null) {
                nameValue = name.getAttributeValue().trim();
            }
        }
        if (nameValue != null) {
            log.debug("Saving local entry with name : " + nameValue);

            if (getSynapseConfiguration().getEntry(nameValue) == null) {
                handleFault(log, "Unable to update local entry. Non existent", null);
            } else {
                entry.setKey(nameValue);
                String value = elem.getAttributeValue(
                    new QName(XMLConfigConstants.NULL_NAMESPACE, "value"));
                String src = elem.getAttributeValue(
                    new QName(XMLConfigConstants.NULL_NAMESPACE, "src"));
                String key = elem.getAttributeValue(
                    new QName(XMLConfigConstants.NULL_NAMESPACE, "key"));
                int type = Integer.parseInt(elem.getAttributeValue(
                    new QName(XMLConfigConstants.NULL_NAMESPACE, "type")));
                OMElement content = elem.getFirstElement();

                if (type == Entry.URL_SRC) {
                    entry.setType(Entry.URL_SRC);
                    try {
                        entry.setSrc(new URL(src));
                    } catch (MalformedURLException e) {
                        handleFault(log, "Invalid URL : '" + src +
                            "' used in local entry definition : " + nameValue, e);
                    }
                } else if (type == Entry.REMOTE_ENTRY) {
                    entry.setType(Entry.REMOTE_ENTRY);
                    entry.setKey(key);
                } else if (type == Entry.INLINE_XML) {
                    entry.setType(Entry.INLINE_XML);
                    entry.setValue(content);
                } else if (type == Entry.INLINE_TEXT) {
                    entry.setType(Entry.INLINE_TEXT);
                    entry.setValue(elem.getText());
                } else {
                    handleFault(log, "Unable to add local entry : " + nameValue +
                        ". Unknown entry type", null);
                }
                getSynapseConfiguration().addEntry(name.getAttributeValue(), entry);
                log.info("Added local entry : " + nameValue + " into the configuration");
            }
        }
        return true;
    }

    /**
     * Returns the OMelement representation of the entry given by sequence
     * name
     *
     * @param entryName - name of the entry to get
     * @return OMElement representing the entryMediator of the given entry
     *         name
     * @throws AxisFault if any error occured while getting the data from the
     *                   SynapseConfiguration
     */
    public OMElement getEntry(String entryName) throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();

        if (synapseConfiguration.getEntry(entryName) != null) {
            OMElement elem = EntrySerializer.serializeEntry(
                synapseConfiguration.getEntryDefinition(entryName), null);
            OMFactory fac = elem.getOMFactory();
            OMNamespace nullNS = fac.createOMNamespace(XMLConfigConstants.NULL_NAMESPACE, "ns3");
            elem.declareNamespace(nullNS);
            if (elem.getAttribute(new QName(XMLConfigConstants.NULL_NAMESPACE, "key")) != null) {
                if (elem.getAttribute(
                    new QName(XMLConfigConstants.NULL_NAMESPACE, "src")) != null) {
                    elem.addAttribute("type", Integer.toString(Entry.URL_SRC), nullNS);
                } else if (elem.getFirstOMChild() instanceof OMText) {
                    elem.addAttribute("type", Integer.toString(
                        Entry.INLINE_TEXT), nullNS);
                } else if (elem.getFirstOMChild() instanceof OMElement) {
                    elem.addAttribute("type", Integer.toString(Entry.INLINE_XML), nullNS);
                }
                return elem;
            } else {
                handleFault(log, "Unable to fetch local entry. Key missing", null);
            }
        } else {
            handleFault(log, "Entry with the key '" + entryName + "' does not exist", null);
        }
        return null;
    }

    /**
     * Deletes the entry with the given name from SynapseConfiguration
     *
     * @param entryName - Name of the entry to delete
     * @throws AxisFault if the entry described by the given name doesnt
     *                   exists in the Synapse Configuration
     */
    public boolean deleteEntry(String entryName) throws AxisFault {
        try {
            log.debug("Deleting local entry with name : " + entryName);
            SynapseConfiguration synapseConfiguration
                = getSynapseConfiguration();
            synapseConfiguration.removeEntry(entryName);
            log.info("Deleted local entry named : " + entryName);
            return true;
        } catch (SynapseException syne) {
            handleFault(log, "Unable to delete the local entry : " + entryName, syne);
        }
        return false;
    }

    /**
     * Returns an String array of the entry names present in the synapse configuration
     *
     * @return String array of entry names
     * @throws AxisFault if an error occurs in getting the synapse configuration
     */
    public String[] getEntryNames() throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        Map gloabalEntriesMap = synapseConfiguration.getLocalRegistry();
        List propKeys = new ArrayList();
        Iterator values = gloabalEntriesMap.values().iterator();
        while (values.hasNext()) {
            Object entryValue = values.next();
            if (entryValue instanceof Entry) {
                propKeys.add(((Entry) entryValue).getKey());
            }
        }
        return (String[]) propKeys.toArray(new String[propKeys.size()]);
    }

    public String getEntryNamesString() throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        Map gloabalEntriesMap = synapseConfiguration.getLocalRegistry();
        List sequenceList = new ArrayList();
        List endpointList = new ArrayList();
        List entryList = new ArrayList();
        StringBuffer entrySb = new StringBuffer();
        StringBuffer endpointSb = new StringBuffer();
        StringBuffer sequenceSb = new StringBuffer();
        Iterator values = gloabalEntriesMap.values().iterator();
        while (values.hasNext()) {
            Object entryValue = values.next();
            if (entryValue instanceof Endpoint) {
                Endpoint endpoint = (Endpoint) entryValue;
                String name = endpoint.getName();
                if (name != null) {
                    endpointList.add(name);
                }
            } else if (entryValue instanceof SequenceMediator) {
                SequenceMediator sequenceMediator = (SequenceMediator) entryValue;
                String name = sequenceMediator.getName();
                if (name != null) {
                    sequenceList.add(name);
                }
            } else if (entryValue instanceof Entry) {
                Entry entry = (Entry) entryValue;
                if (!entry.isDynamic() && !entry.isRemote()) { // only care pre-defined local entries
                    entryList.add(entry.getKey());
                }
            }
        }

        if (!sequenceList.isEmpty()) {
            
            Collections.sort(sequenceList, new Comparator() {
                public int compare(Object o1, Object o2) {
                    return ((String) o1).compareToIgnoreCase(
                        (String) o2);
                }
            });
            for (Iterator it = sequenceList.iterator(); it.hasNext();) {
                String name = (String) it.next();
                if (name != null) {
                    sequenceSb.append("[Sequence]-" + name + " ");
                }
            }
        }

        if (!entryList.isEmpty()) {

            Collections.sort(entryList, new Comparator() {
                public int compare(Object o1, Object o2) {
                    return ((String) o1).compareToIgnoreCase(
                        (String) o2);
                }
            });
            for (Iterator it = entryList.iterator(); it.hasNext();) {
                String name = (String) it.next();
                if (name != null) {
                    entrySb.append("[Entry]-" + name + " ");
                }
            }
        }

        if (!endpointList.isEmpty()) {

            Collections.sort(endpointList, new Comparator() {
                public int compare(Object o1, Object o2) {
                    return ((String) o1).compareToIgnoreCase(
                        (String) o2);
                }
            });
            for (Iterator it = endpointList.iterator(); it.hasNext();) {
                String name = (String) it.next();
                if (name != null) {
                    endpointSb.append("[Enpoint]-" + name + " ");
                }
            }
        }
        return endpointSb.toString() + entrySb.toString() + sequenceSb.toString();
    }
}

