 /*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */  package org.wso2.esb.services;

 import org.apache.axis2.AxisFault;
 import org.apache.commons.logging.Log;
 import org.apache.commons.logging.LogFactory;
 import org.apache.log4j.Level;
 import org.apache.log4j.LogManager;
 import org.apache.log4j.Logger;
 import org.apache.log4j.PropertyConfigurator;
 import org.wso2.esb.services.tos.LogConfigData;
 import org.wso2.esb.ServiceBusConstants;
 import org.wso2.esb.ServiceBusConfiguration;

 import java.io.*;
 import java.util.Enumeration;
 import java.util.HashMap;
 import java.util.Map;
 import java.util.Properties;

 /**
 * This is a POJO for the log configuration
 */
public class LogAdmin extends AbstractESBAdmin {
    private static final Log log = LogFactory.getLog(LogAdmin.class);

    // these are the loggers we give users to change logging levels 
    // from the log settings interface
    private static String LOGGER_ROOT_ID = "loggerRoot";
   
    private static String LOGGER_SERVICE_ID = "loggerService";
    private static String LOGGER_SERVICE_NAME = "SERVICE_LOGGER";
    private static String LOGGER_TRACE_ID = "loggerTrace";
    private static String LOGGER_TRACE_NAME = "TRACE_LOGGER";
    private static Map loggerMap = new HashMap();
    
    // logger map (key=id value=logger name)
    static {
        loggerMap.put(LOGGER_SERVICE_ID, LOGGER_SERVICE_NAME);
        loggerMap.put(LOGGER_TRACE_ID, LOGGER_TRACE_NAME);
    }

    
    /**
     * Get log4j.properties, logger levels
     * read from the file system.
     * 
     * @return 
     * @throws AxisFault
     */
    public LogConfigData getLogConfig() throws AxisFault {
        try {
            log.info("Get Log4J configuration");
            
            // initialized to blank so that going xml will have <x></x>
            // instead of null elements
            String loggerLevelRoot = "";
            String loggerLevelService = "";
            String loggerLevelTrace = "";
            
            String logConfigLocation = getLogConfigLocation();
            
            // read the log4j.properties file
            StringBuffer logConfig = new StringBuffer(1024);
            BufferedReader reader = new BufferedReader(new FileReader(logConfigLocation));
            String line = null;
            while( (line = reader.readLine()) != null){
                line = line.trim();
                logConfig.append(line + "\n");    
            }
            reader.close();

            // find current logger levels
            Logger rootLogger = Logger.getRootLogger();
            if(rootLogger != null) {
                if(rootLogger.getLevel() != null)
                    loggerLevelRoot = rootLogger.getLevel().toString();
            } 
            
            Enumeration loggers = LogManager.getCurrentLoggers();
            while(loggers.hasMoreElements()) {
                Logger logger = (Logger) loggers.nextElement();
                if(LOGGER_SERVICE_NAME.equals(logger.getName())) {
                    if(logger.getLevel() != null)
                        loggerLevelService = logger.getLevel().toString();                    
                } else if(LOGGER_TRACE_NAME.equals(logger.getName())) {
                    if(logger.getLevel() != null)
                        loggerLevelTrace = logger.getLevel().toString();
                }
            }

            // data are log4j.properties file + selected current logger levels
            LogConfigData logConfigData = new LogConfigData();            
            logConfigData.setLoggerLevelRoot(loggerLevelRoot);
            logConfigData.setLoggerLevelService(loggerLevelService);
            logConfigData.setLoggerLevelTrace(loggerLevelTrace);
            logConfigData.setLogText(logConfig.toString());

            return logConfigData;
            
        } catch (IOException e) {
            handleFault(log, "log4j.properties read error:" + e.getMessage(), e);
            return null;
        }
    }

    
    /**
     * Update the running log4j with the given log4j.properties.
     * 
     * @param logConfig log4j.properties contents
     * @throws AxisFault
     */
    public boolean updateLogConfig(String logConfig) throws AxisFault {
        InputStream is = new ByteArrayInputStream(logConfig.getBytes());
        Properties properties = new Properties();
        try {
            properties.load(is);
            
        } catch (IOException e) {
            handleFault(log, "log4j.properties update error:" + e.getMessage(), e);
            return false;
        }
        
        // update runtime log4j
        PropertyConfigurator.configure(properties);
        log.info("Log configuration updated");
        return true;
    }  

    
    /**
     * Save given log4j.properties to default location.
     * 
     * @param logConfig log4j.properties contents
     * @throws AxisFault
     */
    public boolean saveLogConfig(String logConfig) throws AxisFault {
        if(logConfig == null) {
            log.error("saving log configuration called with null");
            return false;
        }
        
        String logConfigurationLocation = getLogConfigLocation();

        try {
            OutputStream out = new java.io.FileOutputStream(new File(logConfigurationLocation));
            out.write(logConfig.getBytes());
            out.close();
            
        } catch (IOException e) {
            handleFault(log, "log4j.properties write error:" + e.getMessage(), e);
            return false;
        }
       
        // update runtime log4j
        PropertyConfigurator.configure(logConfigurationLocation);
        log.info("log4j.properties saved");
        return true;
    }

    
    /**
     * Get log4j.properties location
     * @return
     */
    private String getLogConfigLocation() {
        return ServiceBusConfiguration.getInstance().getEsbHome() + File.separator
            + ServiceBusConstants.ESB_CONF_DIRECTORY + ServiceBusConstants.LOG4J_PROPS;
    }

   
    /**
     * Change logger level in the log4j.properties file.
     * 
     * @param loggerId logger ID from the loggerMap object
     * @param level e.g. INFO, DEBUG
     * @throws AxisFault
     */
    public boolean changeLogConfigLevel(String loggerId, String level) throws AxisFault {
        // find the logger name from id
        log.info("change log level loggerId=" + loggerId + " level=" + level);
        
        // get logger and change level
        Logger logger = null;
        if(LOGGER_ROOT_ID.equals(loggerId)) {
            logger = Logger.getRootLogger();
            
        } else {
            String loggerName = (String) loggerMap.get(loggerId);
            if(loggerName == null) {
                log.error("logger id not found:" + level);
                return false;
            }
            
            logger = Logger.getLogger(loggerName);
        }
        if(logger != null)
            logger.setLevel(Level.toLevel(level));
        
        return true;
    }
 
}