/*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */

package org.wso2.esb.services;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.AxisFault;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.synapse.SynapseConstants;
import org.apache.synapse.SynapseException;
import org.apache.synapse.config.SynapseConfiguration;
import org.apache.synapse.config.xml.ProxyServiceFactory;
import org.apache.synapse.config.xml.ProxyServiceSerializer;
import org.apache.synapse.config.xml.XMLConfigConstants;
import org.apache.synapse.core.axis2.ProxyService;
import org.wso2.esb.services.tos.ProxyData;

import javax.xml.namespace.QName;
import java.util.*;

/**
 * This is a POJO for the Proxy Service admin service
 */
public class ProxyServiceAdmin extends AbstractESBAdmin {

    private static Log log = LogFactory.getLog(ProxyServiceAdmin.class);

    /**
     * Get all the proxy services wrapped in the ProxyData objects as a list for the
     * admin console to view the details of them
     *
     * @return Array of ProxyData objects in the SynapseConfiguraiton
     *
     * @throws AxisFault in case of a failure in getting proxies
     */
    public ProxyData[] proxyData() throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        Collection proxyCollection = synapseConfiguration.getProxyServices();
        ArrayList proxyServiceList = new ArrayList(1);
        for (Iterator ite = proxyCollection.iterator();
             ite.hasNext();) {
            Object obj = ite.next();
            if (obj instanceof ProxyService) {
                ProxyService proxy = (ProxyService) obj;
                ProxyData data = new ProxyData();
                data.setName(proxy.getName());

                if (proxy.getTargetInSequence() != null || proxy.getTargetOutSequence() != null) {
                    data.setTarget(proxy.getTargetInSequence());
                } else {
                    data.setTarget(proxy.getTargetEndpoint());
                }
                data.setRunning(proxy.isRunning());
                if(proxy.getStatisticsState() == SynapseConstants.STATISTICS_ON) {
                    data.setEnableStatistics(true);
                } else {
                    data.setEnableStatistics(false);
                }
                if(proxy.getTraceState() == SynapseConstants.TRACING_ON) {
                    data.setEnableTracing(true);
                } else {
                    data.setEnableTracing(false);
                }
                if(proxy.getWsdlURI() != null ||
                        proxy.getWSDLKey() != null || proxy.getInLineWSDL() != null) {
                    data.setWsdlAvailabel(true);
                } else {
                    data.setWsdlAvailabel(false);
                }
                if(proxy.getServiceLevelPolicies().isEmpty()) {
                    data.setPolicyAvailabel(false);
                } else {
                    data.setPolicyAvailabel(true);
                }
                proxyServiceList.add(data);
            }
        }

        Collections.sort(proxyServiceList, new Comparator() {
            public int compare(Object o1, Object o2) {
                return ((ProxyData) o1).getName().compareToIgnoreCase(
                    ((ProxyData)o2).getName());
            }
        });

        return (ProxyData[]) proxyServiceList
                .toArray(new ProxyData[proxyServiceList.size()]);
    }

    /**
     * Enables statistics for the specified proxy service
     *
     * @param proxyName
     *              name of the proxy service name of which the statistics need to be enabled
     *
     * @throws AxisFault in case of a failure in enabling statistics
     */
    public void enableStatistics(String proxyName) throws AxisFault {

        try {
            getSynapseConfiguration().getProxyService(
                    proxyName).setStatisticsState(SynapseConstants.STATISTICS_ON);
            log.info("Enabled statistics on proxy service : " + proxyName);
        } catch (SynapseException syne) {
            handleFault(log, "Unable to enable statistics for proxy service " + proxyName, syne);
        }
    }

    /**
     * Disables statistics for the specified proxy servivce
     *
     * @param proxyName
     *              name of the proxy service of which statistics need to be disabled
     *
     * @throws AxisFault in case of a failure in disabling statistics
     */
    public void disableStatistics(String proxyName) throws AxisFault {

        try {
            getSynapseConfiguration().getProxyService(
                    proxyName).setStatisticsState(SynapseConstants.STATISTICS_OFF);
            log.info("Disabled statistics on proxy service : " + proxyName);
        } catch (SynapseException syne) {
            handleFault(log, "Unable to disable statistics for proxy service " + proxyName, syne);
        }
    }

    /**
     * Enables tracing for the specified proxy service
     *
     * @param proxyName
     *              name of the the proxy service of which tracing needs to be enabled
     *
     * @throws AxisFault in case of a failure in enabling tracing
     */
    public void enableTracing(String proxyName) throws AxisFault {

        try {
            getSynapseConfiguration().getProxyService(
                    proxyName).setTraceState(SynapseConstants.TRACING_ON);
            log.info("Enabled tracing on proxy service : " + proxyName);
        } catch (SynapseException syne) {
            handleFault(log, "Unable to enable tracing for proxy service " + proxyName, syne);
        }
    }

    /**
     * Disables tracing for the specified proxy service
     *
     * @param proxyName
     *              name of the proxy service of which tracing needs to be disabled
     *
     * @throws AxisFault in case of a failure in disabling tracing
     */
    public void disableTracing(String proxyName) throws AxisFault {

        try {
            getSynapseConfiguration().getProxyService(
                    proxyName).setTraceState(SynapseConstants.TRACING_OFF);
            log.info("Disabled tracing on proxy service : " + proxyName);
        } catch (SynapseException syne) {
            handleFault(log, "Unable to disable tracing for proxy service " + proxyName, syne);
        }
    }

    /**
     * Get the proxy service configuration as an OMElement
     *
     * @param proxyName
     *              name of the proxy service which needs the configuraiton
     *
     * @return OMElement representing the proxy servicve configuration
     *
     * @throws AxisFault in case of failure in getting the configuration
     */
    public OMElement getProxyServiceElement(String proxyName) throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        try {
            if (synapseConfiguration.getProxyService(proxyName) != null) {
                return ProxyServiceSerializer.serializeProxy(null,
                        synapseConfiguration.getProxyService(proxyName));
            } else {
                handleFault(log,"A proxy service named : " + proxyName + " does not exist", null);
            }
        } catch (SynapseException syne) {
            handleFault(log, "Unable to get the proxy service definition for : " + proxyName, syne);
        }
        return null;
    }

    /**
     * Add a proxy service described by the given OMElement
     *
     * @param proxyServiceElement
     *                          configuraiton of the proxy service which needs to be added
     *
     * @throws AxisFault if the element is not an proxy service or if a proxy service with the
     *                   same name exists
     */
    public void addProxyService(OMElement proxyServiceElement) throws AxisFault {
        try {
            if (proxyServiceElement.getQName().getLocalPart()
                    .equals(XMLConfigConstants.PROXY_ELT.getLocalPart())) {
                String proxyName = proxyServiceElement.getAttributeValue(new QName("name"));
                if (getSynapseConfiguration().getProxyService(proxyName) != null) {
                    handleFault(log, "A proxy service named : "
                            + proxyName + " already exists", null);

                } else {
                    ProxyService proxy = ProxyServiceFactory.createProxy(proxyServiceElement);

                    try {
                        proxy.buildAxisService(getSynapseConfiguration(),
                            getAxisConfig());

                        log.info("Added proxy service : " + proxyName);
                        if(!proxy.isStartOnLoad()) {
                            proxy.stop(getSynapseConfiguration());
                        }

                        if (proxy != null) {
                            if (proxy.getTargetInLineInSequence() != null) {
                                proxy.getTargetInLineInSequence().init(getSynapseEnvironment());
                            }
                            if (proxy.getTargetInLineOutSequence() != null) {
                                proxy.getTargetInLineOutSequence().init(getSynapseEnvironment());
                            }
                            if (proxy.getTargetInLineFaultSequence() != null) {
                                proxy.getTargetInLineFaultSequence().init(getSynapseEnvironment());
                            }
                        }

                        getSynapseConfiguration().addProxyService(proxy.getName(), proxy);

                    } catch (AxisFault af) {
                        log.error("Unable to add Proxy service : " + proxy.getName(), af);
                        if (getAxisConfig().getService(proxy.getName()) != null) {
                            try {
                                getAxisConfig().removeService(proxy.getName());
                            } catch (Exception ignore) {}
                        }
                        throw new AxisFault("Unable to add Proxy service : " + proxy.getName(), af);
                    }
                }
            } else {
                handleFault(log, "Unable to add proxy service. Invalid definition", null);
            }
        } catch (SynapseException syne) {
            handleFault(log, "Unable to add proxy service. Invalid definition", syne);
        }
    }

    /**
     * Alter and saves the proxy service to the SynapseConfiguration as specified by the
     * given OMElement configuration
     *
     * @param proxyServiceElement
     *                          configuration of the proxy service which needs to be altered
     *
     * @throws AxisFault if the service not present or the configuration is wrong or
     *                      in case of a failure in building the axis service
     */
    public void saveProxyService(OMElement proxyServiceElement) throws AxisFault {

        try {
            if (proxyServiceElement.getQName().getLocalPart()
                    .equals(XMLConfigConstants.PROXY_ELT.getLocalPart())) {

                String proxyName = proxyServiceElement.getAttributeValue(new QName("name"));

                ProxyService currentProxy = getSynapseConfiguration().getProxyService(proxyName);
                boolean wasRunning = false;
                if (currentProxy == null) {
                    handleFault(log, "Unable to save proxy service : "
                        + proxyName + ". Does not exist", null);

                } else {
                    wasRunning = currentProxy.isRunning();
                    log.debug("Deleting existing proxy service : " + proxyName);
                    deleteProxyService(proxyName);

                    try {
                        log.debug("Adding proxy service : " + proxyName);
                        addProxyService(proxyServiceElement);
                        log.info("Added added proxy service : " + proxyName);

                        if (!wasRunning &&
                            getSynapseConfiguration().getProxyService(proxyName).isRunning()) {
                            getSynapseConfiguration().getProxyService(
                                    proxyName).stop(getSynapseConfiguration());
                        } else if (wasRunning &&
                            !getSynapseConfiguration().getProxyService(proxyName).isRunning()) {
                            getSynapseConfiguration().getProxyService(
                                    proxyName).start(getSynapseConfiguration());
                        }

                        ProxyService proxy = getSynapseConfiguration().getProxyService(proxyName);
                        if (proxy != null) {
                            if (proxy.getTargetInLineInSequence() != null) {
                                proxy.getTargetInLineInSequence().init(getSynapseEnvironment());
                            }
                            if (proxy.getTargetInLineOutSequence() != null) {
                                proxy.getTargetInLineOutSequence().init(getSynapseEnvironment());
                            }
                            if (proxy.getTargetInLineFaultSequence() != null) {
                                proxy.getTargetInLineFaultSequence().init(getSynapseEnvironment());
                            }
                        }
                    } catch (Exception e) {

                        log.error("Unable to save changes made for the proxy service : "
                            + proxyName + ". Restoring the existing proxy..");
                        getSynapseConfiguration().addProxyService(proxyName, currentProxy);
                        currentProxy.buildAxisService(getSynapseConfiguration(),
                                getAxisConfig());
                        if (!wasRunning) {
                            currentProxy.stop(getSynapseConfiguration());
                        } else {
                            currentProxy.start(getSynapseConfiguration());
                        }

                        handleFault(log, "Unable to save changes made for the proxy service : "
                            + proxyName + ". Restored the existing proxy...", e);
                    }
                }



            } else {
                handleFault(log, "Unable to save proxy service. Invalid definition", null);
            }

        } catch (SynapseException syne) {
            handleFault(log, "Unable to save proxy service. Invalid definition", syne);
        }
    }

    /**
     * Deletes a proxy service from the synapse configuration
     *
     * @param proxyServiceName
     *                      name of the proxy service which needs to be deleted
     *
     * @throws AxisFault if the proxy service name given is not existent in the
     *                   synapse configuration
     */
    public void deleteProxyService(String proxyServiceName) throws AxisFault {
        try {
            log.debug("Deleting proxy service : " + proxyServiceName);
            SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
            synapseConfiguration.removeProxyService(proxyServiceName);
            log.info("Proxy service : " + proxyServiceName + " deleted");
        } catch (SynapseException syne) {
            handleFault(log, "Unable to delete proxy service : " + proxyServiceName, syne);
        }
    }

    /**
     * Get the availabel transport names from the AxisConfiguration
     *
     * @return String array of availabel transport names
     *
     * @throws AxisFault if there is an error
     */
    public String[] getAvailabelTransports() throws AxisFault {
        Object[] transports = getAxisConfig().getTransportsIn().keySet().toArray();
        String[] ret = new String[transports.length];
        for (int i = 0; i < transports.length; i++) {
            ret[i] = (String) transports[i];
        }
        return ret;
    }

    /**
     * Get the availabel sequences from the SynapseConfiguration
     *
     * @return String array of availabel sequence names
     *
     * @throws AxisFault if there is an error
     */
    public String[] getAvailabelSequences() throws AxisFault {
        Object[] sequences = getSynapseConfiguration().getDefinedSequences().keySet().toArray();
        String[] ret = new String[sequences.length];
        for (int i = 0; i < sequences.length; i++) {
            ret[i] = (String) sequences[i];
        }
        return ret;
    }

    /**
     * Get the availabel endpoints from the SynapseConfiguration
     *
     * @return String array of availabel endpoint names
     *
     * @throws AxisFault if there is an error
     */
    public String[] getAvailabelEndpoints() throws AxisFault {
        Object[] endpoints = getSynapseConfiguration().getDefinedEndpoints().keySet().toArray();
        String[] ret = new String[endpoints.length];
        for (int i = 0; i < endpoints.length; i++) {
            ret[i] = (String) endpoints[i];
        }
        return ret;
    }

    /**
     * Starts the service specified by the name
     *
     * @param proxyServiceName
     *                      name of the proxy service which needs to be started
     *
     * @throws AxisFault incase of a failure in starting the service
     */
    public void startProxyService(String proxyServiceName) throws AxisFault {
        log.debug("Starting/Re-starting proxy service : " + proxyServiceName);
        getSynapseConfiguration().getProxyService(proxyServiceName)
                .start(getSynapseConfiguration());
        log.info("Started/Re-started proxy service : " + proxyServiceName);
    }

    /**
     * Stops the service specified by the name
     *
     * @param proxyServiceName
     *                      name of the proxy service which needs to be stoped
     *
     * @throws AxisFault in case of a failure in stopping the service
     */
    public void stopProxyService(String proxyServiceName) throws AxisFault {
        log.debug("Stopping proxy service : " + proxyServiceName);
        getSynapseConfiguration().getProxyService(proxyServiceName)
                .stop(getSynapseConfiguration());
        log.info("Stopped proxy service : " + proxyServiceName);
    }

    /**
     * Redeploying service
     * Removes an existing one,Adds a new one
     *
     * @param proxyServiceName name of the proxy service which needs to be redeployed
     * @throws AxisFault in case of a failure in redeploying the service
     */
    public void redeployProxyService(String proxyServiceName) throws AxisFault {
        ProxyService currentProxy = getSynapseConfiguration().getProxyService(proxyServiceName);
        if (currentProxy != null) {
            log.debug("Redeploying proxy service : " + proxyServiceName);
            OMElement proxyElement = ProxyServiceSerializer.serializeProxy(null, currentProxy);
            saveProxyService(proxyElement);
            log.info("Redeployed proxy service : " + proxyServiceName);
        }
    }
}
