/*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */

package org.wso2.esb.services;

import org.apache.axiom.om.OMAttribute;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.xpath.AXIOMXPath;
import org.apache.axis2.AxisFault;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.synapse.Startup;
import org.apache.synapse.SynapseConstants;
import org.apache.synapse.SynapseException;
import org.apache.synapse.config.SynapseConfiguration;
import org.apache.synapse.config.xml.StartupFinder;
import org.apache.synapse.core.SynapseEnvironment;
import org.apache.synapse.startup.quartz.SimpleQuartz;
import org.apache.synapse.startup.quartz.SimpleQuartzFactory;
import org.jaxen.JaxenException;
import org.wso2.esb.services.tos.TaskData;

import javax.xml.namespace.QName;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

/**
 * 
 */
public class TaskAdmin extends AbstractESBAdmin {

    private Log log = LogFactory.getLog(TaskAdmin.class);

    /**
     * Gives the list of Tasks in the SynapseConfiguration as a TaskData transfer objects
     * 
     * @return Array of TaskData objects
     * @throws AxisFault if any error occured in reading the contents from the
     *                      SynapseConfiguration
     */
    public TaskData[] taskData() throws AxisFault {
        SynapseConfiguration synCfg = getSynapseConfiguration();
        ArrayList<TaskData> taskData = new ArrayList<TaskData>();
        // for the moment we only have tasks
        for (Object obj : synCfg.getStartups()) {
            if (obj instanceof SimpleQuartz) {
                TaskData td = new TaskData();
                SimpleQuartz q = (SimpleQuartz) obj;
                td.setName(q.getName());
                td.setClazz(q.getJobClass());
                td.setEnableStatistics(false);
                td.setEnableTracing(false);
                taskData.add(td);
            }
        }

        Collections.sort(taskData, new Comparator() {
            public int compare(Object o1, Object o2) {
                return ((TaskData) o1).getName().compareToIgnoreCase(((TaskData)o2).getName());
            }
        });

        return taskData.toArray(new TaskData[taskData.size()]);
    }

    /**
     * Deletes the task with the given name from SynapseConfiguration
     *
     * @param taskName - Name of the task to delete
     * @throws AxisFault if the Task described by the given name doesnt
     *                   exists in the Synapse Configuration
     */
    public void deleteTask(String taskName) throws AxisFault {
        try {
            log.debug("Deleting task : " + taskName);

            SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
            Startup st = synapseConfiguration.getStartup(taskName);
            if (st != null) {
                st.destroy();
            }
            synapseConfiguration.removeStartup(taskName);
            log.info("Deleted task : " + taskName + " from the configuration");
        } catch (SynapseException syne) {
            handleFault(log, "Error removing sequence : " + taskName, syne);
        }
    }

    /**
     * Returns the OMelement representation of the task given by task name
     *
     * @param taskName - name of the sequence to get
     * @return OMElement representing the Task of the given task name
     * @throws AxisFault if any error occured while getting the data from the
     *                   SynapseConfiguration
     */
    public OMElement getTask(String taskName) throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        try {
            if (synapseConfiguration.getStartup(taskName) != null) {
              OMElement taskElement = StartupFinder.getInstance().serializeStartup(
                    null, synapseConfiguration.getStartup(taskName));
              
              AXIOMXPath xPathLast = new AXIOMXPath("//syn:property");
              xPathLast.addNamespace("syn", SynapseConstants.SYNAPSE_NAMESPACE);
              List<OMElement> selectNodes = xPathLast.selectNodes(taskElement);
              
              for(OMElement selectedNode : selectNodes) {
                OMAttribute attributeValue = selectedNode.getAttribute(new QName("value"));
                if(attributeValue == null) {
                  OMElement firstElement = selectedNode.getFirstElement();
                  selectedNode.setText(firstElement.toString());
                  firstElement.detach();
                }
                
              }
              
              return taskElement;
            } else {
                handleFault(log, "Task with the name " + taskName + " does not exist", null);
            }
        } catch (SynapseException syne) {
            handleFault(log, "Unable to get the task : " + taskName, syne);
        } catch (JaxenException e) {
            handleFault(log, "Unable to get the task : " + taskName, e);
        }
        return null;
    }

    public void addTask(OMElement taskElement) throws AxisFault {

        SimpleQuartzFactory fac = new SimpleQuartzFactory();
        Startup st = fac.createStartup(taskElement.getFirstElement());
        getSynapseConfiguration().addStartup(st);
        Object o = getAxisConfig().getParameter(SynapseConstants.SYNAPSE_ENV).getValue();
        if (o instanceof SynapseEnvironment) {
            st.init((SynapseEnvironment) o);
        } else {
            handleFault(log, "Can not initialize the task. Task is added but not scheduled.", null);
        }
    }

    public void editTask(OMElement taskElement) throws AxisFault {

        SimpleQuartzFactory fac = new SimpleQuartzFactory();
        Startup st = fac.createStartup(taskElement.getFirstElement());
        deleteTask(st.getName());
        getSynapseConfiguration().addStartup(st);
        Object o = getAxisConfig().getParameter(SynapseConstants.SYNAPSE_ENV).getValue();
        if (o instanceof SynapseEnvironment) {
            st.init((SynapseEnvironment) o);
        } else {
            handleFault(log, "Can not initialize the task. Task is added but not scheduled.", null);
        }
    }
}
