/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.esb.mediators.transform;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.synapse.MessageContext;
import org.apache.synapse.mediators.AbstractMediator;
import org.apache.synapse.util.PayloadHelper;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.milyn.Smooks;
import org.w3c.dom.Document;

import javax.xml.stream.XMLStreamException;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.stream.StreamSource;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

/**
 * Transforms the current message payload using the given Smooks configuration.
 * The current message context is replaced with the result as XML.
 */
public class SmooksMediator extends AbstractMediator {

    /**
     * Smooks engine
     */
    Smooks smooks = null;
    /**
     * Smooks configuration file
     */
    String smooksConfig = null;

    public boolean mediate(MessageContext synCtx) {
        boolean traceOn = isTraceOn(synCtx);
        boolean traceOrDebugOn = isTraceOrDebugOn(traceOn);

        if (traceOrDebugOn) {
            traceOrDebug(traceOn, "Start : SmooksTransform mediator");

            if (traceOn && trace.isTraceEnabled()) {
                trace.trace("Message : " + synCtx.getEnvelope());
            }
        }

        ByteArrayInputStream byteArrayInputStream = null;

        int payloadType = PayloadHelper.getPayloadType(synCtx);
        if (payloadType == PayloadHelper.XMLPAYLOADTYPE) {
            OMElement payload = PayloadHelper.getXMLPayload(synCtx.getEnvelope());
            try {
                byteArrayInputStream = new ByteArrayInputStream(
                        payload.toStringWithConsume().getBytes());

            } catch (XMLStreamException e) {
                handleException("Cannot read XML payload", e, synCtx);
            }

        } else if (payloadType == PayloadHelper.TEXTPAYLOADTYPE) {
            String payload = PayloadHelper.getTextPayload(synCtx);
            byteArrayInputStream = new ByteArrayInputStream(payload.getBytes());

        } else {
            handleException("Payload is not TEXT, XML", synCtx);
        }

        org.milyn.container.ExecutionContext executionContext = smooks.createExecutionContext();
        DOMResult domResult = new DOMResult();
        smooks.filter(new StreamSource(byteArrayInputStream), domResult, executionContext);

        StAXOMBuilder builder = null;
        try {
            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
            Document doc = (Document) domResult.getNode();
            OutputFormat format = new OutputFormat(doc);
            format.setIndenting(true);
            XMLSerializer serializer = new XMLSerializer(outputStream, format);
            serializer.serialize(doc);

            ByteArrayInputStream inputStream = new ByteArrayInputStream(outputStream.toByteArray());
            builder = new StAXOMBuilder(inputStream);

        } catch (XMLStreamException e) {
            handleException("DOM to Axiom error", e, synCtx);
        } catch (FileNotFoundException e) {
            handleException("DOM to Axiom error", e, synCtx);
        } catch (IOException e) {
            handleException("DOM to Axiom error", e, synCtx);
        }
        OMElement transformedElement = builder.getDocumentElement();
        PayloadHelper.setXMLPayload(synCtx, transformedElement);

        if (traceOrDebugOn) {
            traceOrDebug(traceOn, "Transform completed");

            if (traceOn && trace.isTraceEnabled()) {
                trace.trace("Transformed message : " + synCtx.getEnvelope());
            }
        }

        if (traceOrDebugOn) {
            traceOrDebug(traceOn, "End : SmooksTransform mediator");
        }

        return true;
    }

    public Smooks getSmooks() {
        return smooks;
    }

    public void setSmooks(Smooks smooks) {
        this.smooks = smooks;
    }

    public String getSmooksConfig() {
        return smooksConfig;
    }

    public void setSmooksConfig(String smooksConfig) {
        this.smooksConfig = smooksConfig;
    }
}
