/*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */

package org.wso2.esb;

import org.apache.catalina.Context;
import org.apache.catalina.Engine;
import org.apache.catalina.Host;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.connector.Connector;
import org.apache.catalina.realm.MemoryRealm;
import org.apache.catalina.startup.Embedded;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.tomcat.util.IntrospectionUtils;


import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.util.Properties;

/**
 * The embedded Tomcat server that will host the ESB console
 */

public class TomcatServer {

    private static Log log = LogFactory.getLog(TomcatServer.class);
    private Embedded embedded;
    private Properties props = null;

    public TomcatServer() {
        props = new Properties();
        try {
            props.load(this.getClass().getClassLoader().getResourceAsStream("tomcat.properties"));
        } catch (IOException ignore) {}
        embedded = new Embedded();
    }

    /**
     * Start the embedded Tomcat server
     * @throws Exception on error
     */
    public void start() throws Exception {

        log.info("Starting the embedded Tomcat server..");

        embedded.setCatalinaHome(
            new File(".").getAbsolutePath() + File.separator + "tomcat");

        String webappsDir = "webapp";
        String docsDir    = "docs";
        String samplesDir = "samples";

        // set the memory realm
        MemoryRealm memRealm = new MemoryRealm();
        embedded.setRealm(memRealm);

        // Create an engine
        String host = InetAddress.getLocalHost().getHostName();
        Engine engine = embedded.createEngine();
        engine.setName("Catalina");
        engine.setDefaultHost(host);
        embedded.addEngine(engine);

        // Create a default virtual host
        Host defaultHost = embedded.createHost(host, webappsDir);
        engine.addChild(defaultHost);
        String esbContextPath = getTomcatProperty("ContextPath", "/esb");

        Context esbContext = embedded.createContext(esbContextPath, new File(webappsDir).getAbsolutePath());
        defaultHost.addChild(esbContext);

        Context docsContext = embedded.createContext("/docs", new File(docsDir).getAbsolutePath());
        defaultHost.addChild(docsContext);

        Context samplesContext = embedded.createContext("/samples", new File(samplesDir).getAbsolutePath());
        defaultHost.addChild(samplesContext);

        defaultHost.setDeployOnStartup(true);
        embedded.setUseNaming(true);

        // add the ssl listner for https transport - admin console
        addSSLConnector();

        // Start the embedded server
        try {
            embedded.start();
            log.info("Tomcat Server started at " + "https://" + host + ":"
                + getTomcatProperty("ConsolePort", "9444") + esbContextPath);
        } catch (Exception e) {
            log.fatal("Error starting embedded Tomcat server : ", e);
            throw new ServiceBusException(e);
        }
    }

    private void addSSLConnector() {

        String ksLocation = getTomcatProperty("KeyStore.Location", "webapp/WEB-INF/classes/conf/identity.jks");
        String ksPassword = getTomcatProperty("KeyStore.Password", "password");

        String absKsLocation;
        if (!new File(ksLocation).isAbsolute()) {
            absKsLocation = new File(ksLocation).getAbsolutePath();
        } else {
            absKsLocation = ksLocation;
        }

        String sslProtocol          = getTomcatProperty("SslProtocol");
        String maxHttpHeaderSize    = getTomcatProperty("MaxHttpHeaderSize");
        String maxThreads           = getTomcatProperty("MaxThreads");
        String minSpareThreads      = getTomcatProperty("MinSpareThreads");
        String maxSpareThreads      = getTomcatProperty("MaxSpareThreads");
        String clientAuth           = getTomcatProperty("ClientAuth");
        String disableUploadTimeout = getTomcatProperty("DisableUploadTimeout");
        String acceptCount          = getTomcatProperty("AcceptCount");

        // SSLConnector
        Connector sslConnector = embedded.createConnector(
            (InetAddress) null, Integer.parseInt(getTomcatProperty("ConsolePort", "9444")), true);
        sslConnector.setScheme(ServiceBusConstants.HTTPS_TRANSPORT);

        if (ksPassword != null) {
            IntrospectionUtils.setProperty(sslConnector, "keypass", ksPassword);
        }
        if (ksLocation != null) {
            IntrospectionUtils.setProperty(sslConnector, "keystore", absKsLocation);
        }
        if (sslProtocol != null) {
            IntrospectionUtils.setProperty(sslConnector, "sslProtocol", sslProtocol);
        }
        if (maxHttpHeaderSize != null) {
            IntrospectionUtils.setProperty(sslConnector, "maxHttpHeaderSize", maxHttpHeaderSize);
        }
        if (maxThreads != null) {
            IntrospectionUtils.setProperty(sslConnector, "maxThreads", maxThreads);
        }
        if (maxSpareThreads != null) {
            IntrospectionUtils.setProperty(sslConnector, "maxSpareThreads", maxSpareThreads);
        }
        if (minSpareThreads != null) {
            IntrospectionUtils.setProperty(sslConnector, "minSpareThreads", minSpareThreads);
        }
        if (disableUploadTimeout != null) {
            IntrospectionUtils.setProperty(sslConnector, "disableUploadTimeout", disableUploadTimeout);
        }
        if (clientAuth != null) {
            IntrospectionUtils.setProperty(sslConnector, "clientAuth", clientAuth);
        }
        if (acceptCount != null) {
            IntrospectionUtils.setProperty(sslConnector, "acceptCount", acceptCount);
        }

        sslConnector.setEnableLookups(true);
        sslConnector.setSecure(true);
        embedded.addConnector(sslConnector);
    }

    /**
     * Stop the Tomcat server
     */
    public void stop() {
        log.info("Stopping the embedded Tomcat server..");
        try {
            embedded.stop();
        } catch (LifecycleException e) {
            log.error("Error occurred while stopping the embedded Tomcat server", e);
        }
    }

    private String getTomcatProperty(String key, String def) {
        String res = props.getProperty(key);
        if (res == null) {
            return def;
        } else {
            return res;
        }
    }

    private String getTomcatProperty(String key) {
        return props.getProperty(key);
    }
}
